import random
from enum import Enum

import torch
from torch.utils.data import DataLoader, Dataset

global_rng = random.Random()


def ids_tensor(shape, vocab_size, rng=None, name=None):
    """Creates a random int32 tensor of the shape within the vocab size."""
    if rng is None:
        rng = global_rng

    total_dims = 1
    for dim in shape:
        total_dims *= dim

    values = []
    for _ in range(total_dims):
        values.append(rng.randint(0, vocab_size - 1))

    return torch.tensor(data=values, dtype=torch.long).view(shape).contiguous()


def floats_tensor(shape, scale=1.0, rng=None, name=None):
    """Creates a random float32 tensor of the shape within the vocab size."""
    if rng is None:
        rng = global_rng

    total_dims = 1
    for dim in shape:
        total_dims *= dim

    values = []
    for _ in range(total_dims):
        values.append(rng.random() * scale)

    return torch.tensor(data=values, dtype=torch.float).view(shape).contiguous()


def generate_sample(desc, device=None):
    """Generate a sample based on the description"""
    # symbolic dimensions are described with strings. set symbolic dimensions to be 1
    size = [s if isinstance(s, (int)) else 1 for s in desc.shape_]
    if desc.num_classes_:
        return torch.randint(0, desc.num_classes_, size, dtype=desc.dtype_).to(device)
    else:
        return torch.randn(size, dtype=desc.dtype_).to(device)


class OrtTestDataset(Dataset):
    def __init__(self, input_desc, seq_len, dataset_len, device):
        import copy

        self.input_desc_ = copy.deepcopy(input_desc)
        for input_desc in self.input_desc_:
            shape_ = []
            for i, axis in enumerate(input_desc.shape_):
                if axis == "max_seq_len_in_batch":
                    shape_ = [*shape_, seq_len]
                elif axis != "batch":
                    shape_ = input_desc.shape_[i]
            input_desc.shape_ = shape_
        self.dataset_len_ = dataset_len
        self.device_ = device

    def __len__(self):
        return self.dataset_len_

    def __getitem__(self, item):
        input_batch = []
        for input_desc in self.input_desc_:
            input_sample = generate_sample(input_desc, self.device_)
            input_batch.append(input_sample)
        return input_batch


def create_ort_test_dataloader(input_desc, batch_size, seq_len, dataset_len, device):
    dataset = OrtTestDataset(input_desc, seq_len, dataset_len, device)
    return DataLoader(dataset, batch_size=batch_size)


class BatchArgsOption(Enum):
    List = 1
    Dict = 2
    ListAndDict = 3


def split_batch(batch, input_desc, args_count):
    total_argument_count = len(input_desc)
    # batch=[input_ids[batch, seglen], attention_mask[batch, seglen], token_type_ids[batch,seglen], token_type_ids[batch, seglen]]
    args = []  # (input_ids[batch, seglen], attention_mask[batch, seglen])
    kwargs = {}  # {'token_type_ids': token_type_ids[batch,seglen], 'position_ids': token_type_ids[batch, seglen]}
    for i in range(args_count):
        args = [*args, batch[i]]

    for i in range(args_count, total_argument_count):
        kwargs[input_desc[i].name_] = batch[i]

    return args, kwargs



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import os
import subprocess
import sys
from collections import namedtuple

SCRIPT_DIR = os.path.realpath(os.path.dirname(__file__))


def parse_args():
    parser = argparse.ArgumentParser(description="Runs GPT-2 performance tests.")
    parser.add_argument("--binary_dir", required=True, help="Path to the ORT binary directory.")
    parser.add_argument("--training_data_root", required=True, help="Path to the training data root directory.")
    parser.add_argument("--model_root", required=True, help="Path to the model root directory.")
    return parser.parse_args()


# TODO - review to finalize params
def main():
    args = parse_args()

    Config = namedtuple("Config", ["use_mixed_precision", "max_seq_length", "batch_size"])
    configs = [Config(True, 1024, 1), Config(False, 1024, 1)]

    # run GPT-2 training
    for c in configs:
        print(
            "######## testing name - "
            + ("fp16-" if c.use_mixed_precision else "fp32-")
            + str(c.max_seq_length)
            + " ##############"
        )
        cmds = [
            os.path.join(args.binary_dir, "onnxruntime_training_gpt2"),
            "--model_name",
            os.path.join(
                args.model_root,
                "megatron-gpt2_hidden-size-1024_num-layers-24_vocab-size-50257_num-attention-heads-16_max-position-embeddings-1024_optimized_opset12",
            ),
            "--train_data_dir",
            os.path.join(args.training_data_root, "train"),
            "--test_data_dir",
            os.path.join(args.training_data_root, "test"),
            "--train_batch_size",
            str(c.batch_size),
            "--mode",
            "train",
            "--max_seq_length",
            str(c.max_seq_length),
            "--num_train_steps",
            "640",
            "--gradient_accumulation_steps",
            "1",
            "--perf_output_dir",
            os.path.join(SCRIPT_DIR, "results"),
        ]

        if c.use_mixed_precision:
            (cmds.append("--use_mixed_precision"),)

        subprocess.run(cmds).check_returncode()  # noqa: PLW1510

    return 0


if __name__ == "__main__":
    sys.exit(main())



import onnx
from onnx import OperatorSetIdProto, TensorProto, helper

# inputs and outputs
hidden = 1024
head = 16
inputs = [
    helper.make_tensor_value_info("input", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden]),
    helper.make_tensor_value_info("attention_mask", TensorProto.INT64, ["batch_size", "sequence_length"]),
    helper.make_tensor_value_info("matmul1.weight", TensorProto.FLOAT16, [hidden, 1024]),
    helper.make_tensor_value_info("add1.bias", TensorProto.FLOAT16, [hidden]),
    helper.make_tensor_value_info("matmul2.weight", TensorProto.FLOAT16, [hidden, 1024]),
    helper.make_tensor_value_info("add2.bias", TensorProto.FLOAT16, [hidden]),
    helper.make_tensor_value_info("matmul3.weight", TensorProto.FLOAT16, [hidden, 1024]),
    helper.make_tensor_value_info("add3.bias", TensorProto.FLOAT16, [hidden]),
    helper.make_tensor_value_info("matmul4.weight", TensorProto.FLOAT16, [hidden, 1024]),
    helper.make_tensor_value_info("add4.bias", TensorProto.FLOAT16, [hidden]),
    helper.make_tensor_value_info("layer_norm1.weight", TensorProto.FLOAT, [hidden]),
    helper.make_tensor_value_info("layer_norm1.bias", TensorProto.FLOAT, [hidden]),
    helper.make_tensor_value_info("matmul7.weight", TensorProto.FLOAT16, [hidden, hidden * 4]),
    helper.make_tensor_value_info("add7.bias", TensorProto.FLOAT16, [hidden * 4]),
    helper.make_tensor_value_info("matmul8.weight", TensorProto.FLOAT16, [hidden * 4, hidden]),
    helper.make_tensor_value_info("add8.bias", TensorProto.FLOAT16, [hidden]),
    helper.make_tensor_value_info("layer_norm2.weight", TensorProto.FLOAT, [hidden]),
    helper.make_tensor_value_info("layer_norm2.bias", TensorProto.FLOAT, [hidden]),
]

outputs = [
    helper.make_tensor_value_info("final_output", TensorProto.FLOAT, ["batch_size", hidden]),
]

# initializers

initializers = [
    helper.make_tensor("scalar_float_0.1", TensorProto.FLOAT, [], [0.1]),
    helper.make_tensor("scalar_float_0", TensorProto.FLOAT, [], [0.0]),
    helper.make_tensor("scalar_float16_8", TensorProto.FLOAT16, [], [8]),
    helper.make_tensor("scalar_bool_true", TensorProto.BOOL, [], [1]),
    helper.make_tensor("scalar_float_1", TensorProto.FLOAT, [], [1]),
    helper.make_tensor("scalar_float_big_num", TensorProto.FLOAT, [], [-3.4028234663852886e38]),
    helper.make_tensor("scalar_int_0", TensorProto.INT64, [], [0]),
    helper.make_tensor("scalar_int_1", TensorProto.INT64, [], [1]),
    helper.make_tensor("single_value_1d_int_0", TensorProto.INT64, [1], [0]),
    helper.make_tensor("single_value_1d_int_1", TensorProto.INT64, [1], [1]),
    helper.make_tensor("single_value_1d_int_2", TensorProto.INT64, [1], [2]),
    helper.make_tensor("single_value_1d_int_16", TensorProto.INT64, [1], [head]),
    helper.make_tensor("single_value_1d_int_64", TensorProto.INT64, [1], [hidden // head]),
    helper.make_tensor("single_value_1d_int_1024", TensorProto.INT64, [1], [hidden]),
    helper.make_tensor("shape1", TensorProto.INT64, [4], [0, 0, 16, 64]),
    helper.make_tensor("shape2", TensorProto.INT64, [4], [0, 0, 16, 64]),
    helper.make_tensor("shape3", TensorProto.INT64, [4], [0, 0, 16, 64]),
    helper.make_tensor("shape4", TensorProto.INT64, [3], [0, 0, 1024]),
]

# nodes

nodes = [
    helper.make_node("Dropout", ["input", "scalar_float_0", "scalar_bool_true"], ["d1_out", "d1_mask"], "d1"),
    helper.make_node("Cast", ["d1_out"], ["c1_out"], name="c1", to=10),
    # attention
    ## left branch
    helper.make_node("MatMul", ["c1_out", "matmul1.weight"], ["m1_out"], "m1"),
    helper.make_node("Add", ["add1.bias", "m1_out"], ["a1_out"], "a1"),
    helper.make_node("Reshape", ["a1_out", "shape1"], ["reshape1_out"], "reshape1"),
    helper.make_node("Transpose", ["reshape1_out"], ["transpose1_out"], name="transpose1", perm=[0, 2, 1, 3]),
    ## middle branch
    helper.make_node("MatMul", ["c1_out", "matmul2.weight"], ["m2_out"], "m2"),
    helper.make_node("Add", ["add2.bias", "m2_out"], ["a2_out"], "a2"),
    helper.make_node("Reshape", ["a2_out", "shape2"], ["reshape2_out"], "reshape2"),
    helper.make_node("Transpose", ["reshape2_out"], ["transpose2_out"], name="transpose2", perm=[0, 2, 1, 3]),
    ## right banch
    helper.make_node("MatMul", ["c1_out", "matmul3.weight"], ["m3_out"], "m3"),
    helper.make_node("Add", ["add3.bias", "m3_out"], ["a3_out"], "a3"),
    helper.make_node("Reshape", ["a3_out", "shape3"], ["reshape3_out"], "reshape3"),
    helper.make_node("Transpose", ["reshape3_out"], ["transpose3_out"], name="transpose3", perm=[0, 2, 3, 1]),
    ## middle branch result computes with right branch result
    helper.make_node("MatMul", ["transpose2_out", "transpose3_out"], ["m4_out"], "m4"),
    helper.make_node("Div", ["m4_out", "scalar_float16_8"], ["div1_out"], "div1"),
    helper.make_node("Cast", ["div1_out"], ["c2_out"], name="c2", to=1),
    helper.make_node("Unsqueeze", ["attention_mask", "single_value_1d_int_1"], ["unsqueeze7_out"], "unsqueeze7"),
    helper.make_node("Unsqueeze", ["unsqueeze7_out", "single_value_1d_int_2"], ["unsqueeze8_out"], "unsqueeze8"),
    helper.make_node("Cast", ["unsqueeze8_out"], ["c3_out"], name="c3", to=1),
    helper.make_node("Sub", ["scalar_float_1", "c3_out"], ["sub1_out"], "sub1"),
    helper.make_node("Mul", ["sub1_out", "scalar_float_big_num"], ["mul1_out"], "mul1"),
    helper.make_node("Add", ["mul1_out", "c2_out"], ["a4_out"], "a4"),
    helper.make_node("Softmax", ["a4_out"], ["softmax1_out"], "softmax1", axis=-1),
    helper.make_node("Dropout", ["softmax1_out", "scalar_float_0", "scalar_bool_true"], ["d2_out", "d2_mask"], "d2"),
    helper.make_node("Cast", ["d2_out"], ["c4_out"], name="c4", to=10),
    ## left branch result computes with result of `middle branch result computes with right branch result``
    helper.make_node("MatMul", ["c4_out", "transpose1_out"], ["m5_out"], "m5"),
    helper.make_node("Transpose", ["m5_out"], ["tranpose4_out"], name="tranpose4", perm=[0, 2, 1, 3]),
    helper.make_node("Reshape", ["tranpose4_out", "shape4"], ["reshape4_out"], "reshape4"),
    ## attention output
    helper.make_node("MatMul", ["reshape4_out", "matmul4.weight"], ["m6_out"], "m6"),
    helper.make_node("Add", ["add4.bias", "m6_out"], ["a5_out"], "a5"),
    helper.make_node("Dropout", ["a5_out", "scalar_float_0", "scalar_bool_true"], ["d4_out", 'd4_mask"'], "d4"),
    helper.make_node("Cast", ["d4_out"], ["c5_out"], name="c5", to=1),
    helper.make_node("Add", ["d1_out", "c5_out"], ["a6_out"], "a6"),
    # MLP
    helper.make_node(
        "LayerNormalization",
        ["a6_out", "layer_norm1.weight", "layer_norm1.bias"],
        ["layernorm1_out", "layernorm1_mean", "layernorm1_var"],
        "layernorm1",
        axis=-1,
        epsion=0.000009999999747378752,
    ),
    helper.make_node("Cast", ["layernorm1_out"], ["c6_out"], name="c6", to=10),
    helper.make_node("MatMul", ["c6_out", "matmul7.weight"], ["m7_out"], "m7"),
    helper.make_node("BiasGelu", ["m7_out", "add7.bias"], ["biasgelu1_out"], "biasgelu1", domain="com.microsoft"),
    helper.make_node("MatMul", ["biasgelu1_out", "matmul8.weight"], ["m8_out"], "m8"),
    helper.make_node("Add", ["add8.bias", "m8_out"], ["a7_out"], "a7"),
    helper.make_node("Dropout", ["a7_out", "scalar_float_0", "scalar_bool_true"], ["d5_out", "d5_mask"], "d5"),
    helper.make_node("Cast", ["d5_out"], ["c7_out"], name="c7", to=1),
    helper.make_node("Add", ["layernorm1_out", "c7_out"], ["a8_out"], "a8"),
    helper.make_node(
        "LayerNormalization",
        ["a8_out", "layer_norm2.weight", "layer_norm2.bias"],
        ["layernorm2_out", "layernorm2_mean", "layernorm2_var"],
        "layernorm2",
        axis=-1,
        epsion=0.000009999999747378752,
    ),
    helper.make_node("Gather", ["layernorm2_out", "scalar_int_0"], ["final_gather_out"], "final_gather", axis=1),
    helper.make_node(
        "Dropout", ["final_gather_out", "scalar_float_0", "scalar_bool_true"], ["final_output", "d6_mask"], "d6"
    ),
]


# Shapes that cannot be inferred by onnx shape inference
value_infos = [
    helper.make_value_info(
        name="reshape1_out",
        type_proto=helper.make_tensor_type_proto(
            elem_type=TensorProto.FLOAT16, shape=["batch_size", "sequence_length", head, hidden // head]
        ),
    ),
    helper.make_value_info(
        name="reshape2_out",
        type_proto=helper.make_tensor_type_proto(
            elem_type=TensorProto.FLOAT16, shape=["batch_size", "sequence_length", head, hidden // head]
        ),
    ),
    helper.make_value_info(
        name="reshape3_out",
        type_proto=helper.make_tensor_type_proto(
            elem_type=TensorProto.FLOAT16, shape=["batch_size", "sequence_length", head, hidden // head]
        ),
    ),
    helper.make_value_info(
        name="reshape4_out",
        type_proto=helper.make_tensor_type_proto(
            elem_type=TensorProto.FLOAT16, shape=["batch_size", "sequence_length", hidden]
        ),
    ),
    helper.make_value_info(
        name="layernorm1_out",
        type_proto=helper.make_tensor_type_proto(
            elem_type=TensorProto.FLOAT, shape=["batch_size", "sequence_length", hidden]
        ),
    ),
    helper.make_value_info(
        name="layernorm2_out",
        type_proto=helper.make_tensor_type_proto(
            elem_type=TensorProto.FLOAT, shape=["batch_size", "sequence_length", hidden]
        ),
    ),
    helper.make_value_info(
        name="concattraining4_out",
        type_proto=helper.make_tensor_type_proto(elem_type=TensorProto.INT64, shape=[3]),
    ),
    helper.make_value_info(
        name="biasgelu1_out",
        type_proto=helper.make_tensor_type_proto(
            elem_type=TensorProto.FLOAT16, shape=["batch_size", "sequence_length", hidden * 4]
        ),
    ),
]

# Create the graph (GraphProto)
graph_def = helper.make_graph(
    nodes,
    "test-model",
    inputs,
    outputs,
    initializers,
    "doc string",
    value_infos,
)


opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 14
onnxdomain.domain = ""  # The empty string ("") or absence of this field implies the operator
# set that is defined as part of the ONNX specification.
opsets.append(onnxdomain)

msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"

opsets.append(msdomain)
kwargs = {}
kwargs["opset_imports"] = opsets


model_def = helper.make_model(graph_def, producer_name="onnx-example", **kwargs)
final_model = onnx.shape_inference.infer_shapes(model_def)
onnx.save(final_model, "gather_roberta_e2e.onnx")



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import unittest

import numpy as np
import onnx
from onnx import TensorProto, helper
from op_test_utils import TestDataFeeds, check_model_correctness, check_op_type_count

from onnxruntime.quantization import quantize_dynamic


class TestOpAttention(unittest.TestCase):
    def input_feeds(self, n, name2shape):
        input_data_list = []
        for _i in range(n):
            inputs = {}
            for name, shape in name2shape.items():
                inputs.update({name: np.random.randint(-1, 2, shape).astype(np.float32)})
            input_data_list.extend([inputs])
        dr = TestDataFeeds(input_data_list)
        return dr

    def construct_model_attention_and_matmul(self, output_model_path):
        #      (input)
        #         |
        #     Attention
        #         |
        #       MatMul
        #         |
        #      (output)
        input_name = "input"
        output_name = "output"
        initializers = []

        def make_attention_node(input_name, weight_shape, weight_name, bias_shape, bias_name, output_name):
            weight_data = np.random.normal(0, 0.1, weight_shape).astype(np.float32)
            initializers.append(onnx.numpy_helper.from_array(weight_data, name=weight_name))

            bias_data = np.random.normal(0, 0.1, bias_shape).astype(np.float32)
            initializers.append(onnx.numpy_helper.from_array(bias_data, name=bias_name))

            return onnx.helper.make_node("Attention", [input_name, weight_name, bias_name], [output_name])

        def make_matmul_node(input_name, weight_shape, weight_name, output_name):
            weight_data = np.random.normal(0, 0.1, weight_shape).astype(np.float32)
            initializers.append(onnx.numpy_helper.from_array(weight_data, name=weight_name))

            return onnx.helper.make_node("MatMul", [input_name, weight_name], [output_name])

        # make attention node
        attention_output_name = "attention_output"
        attention_node = make_attention_node(
            input_name, [10, 30], "qkv.weight", [30], "qkv.bias", attention_output_name
        )
        attention_node.domain = "com.microsoft"
        attention_node.attribute.extend([helper.make_attribute("num_heads", 5)])

        # make matmul node
        matmul_node = make_matmul_node(attention_output_name, [10, 10], "matmul.weight", output_name)

        # make graph
        input_tensor = helper.make_tensor_value_info(input_name, TensorProto.FLOAT, [1, -1, 10])
        output_tensor = helper.make_tensor_value_info(output_name, TensorProto.FLOAT, [1, -1, 10])
        graph_name = "attention_test"
        graph = helper.make_graph(
            [attention_node, matmul_node],
            graph_name,
            [input_tensor],
            [output_tensor],
            initializer=initializers,
        )
        model = helper.make_model(
            graph, opset_imports=[helper.make_opsetid("", 13), helper.make_opsetid("com.microsoft", 1)]
        )
        model.ir_version = onnx.IR_VERSION

        onnx.save(model, output_model_path)

    def dynamic_attention_quant_test(self, model_fp32_path, model_int8_path, per_channel, reduce_range):
        quantize_dynamic(
            model_fp32_path,
            model_int8_path,
            per_channel=per_channel,
            reduce_range=reduce_range,
        )
        quant_nodes = {"QAttention": 1, "MatMulInteger": 1}
        check_op_type_count(self, model_int8_path, **quant_nodes)
        check_model_correctness(
            self,
            model_fp32_path,
            model_int8_path,
            {"input": np.random.rand(1, 5, 10).astype(np.float32)},
        )

    def test_quantize_attention(self):
        np.random.seed(1)
        model_fp32_path = "attention_fp32.onnx"
        model_int8_path = "attention_fp32.quant.onnx"
        self.construct_model_attention_and_matmul(model_fp32_path)

        self.dynamic_attention_quant_test(model_fp32_path, model_int8_path, True, True)
        self.dynamic_attention_quant_test(model_fp32_path, model_int8_path, True, False)
        self.dynamic_attention_quant_test(model_fp32_path, model_int8_path, False, True)
        self.dynamic_attention_quant_test(model_fp32_path, model_int8_path, False, False)


if __name__ == "__main__":
    unittest.main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class FloatProperty(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = FloatProperty()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsFloatProperty(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def FloatPropertyBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x4F\x44\x54\x43", size_prefixed=size_prefixed)

    # FloatProperty
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # FloatProperty
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # FloatProperty
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def FloatPropertyStart(builder):
    builder.StartObject(2)

def Start(builder):
    FloatPropertyStart(builder)

def FloatPropertyAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    FloatPropertyAddName(builder, name)

def FloatPropertyAddValue(builder, value):
    builder.PrependFloat32Slot(1, value, 0.0)

def AddValue(builder, value):
    FloatPropertyAddValue(builder, value)

def FloatPropertyEnd(builder):
    return builder.EndObject()

def End(builder):
    return FloatPropertyEnd(builder)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

import numpy as np
from helper import get_name
from onnx import load

import onnxruntime.backend as backend
from onnxruntime import datasets
from onnxruntime.backend.backend import OnnxRuntimeBackend as ort_backend


def check_list_of_map_to_float(testcase, expected_rows, actual_rows):
    """Validate two list<map<key, float>> instances match closely enough."""

    num_rows = len(expected_rows)
    sorted_keys = sorted(expected_rows[0].keys())
    testcase.assertEqual(num_rows, len(actual_rows))
    testcase.assertEqual(sorted_keys, sorted(actual_rows[0].keys()))

    for i in range(num_rows):
        # use np.testing.assert_allclose so we can specify the tolerance
        np.testing.assert_allclose(
            [expected_rows[i][key] for key in sorted_keys],
            [actual_rows[i][key] for key in sorted_keys],
            rtol=1e-05,
            atol=1e-07,
        )


class TestBackend(unittest.TestCase):
    def test_run_model_non_tensor(self):
        name = get_name("pipeline_vectorize.onnx")
        rep = backend.prepare(name)
        x = {0: 25.0, 1: 5.13, 2: 0.0, 3: 0.453, 4: 5.966}
        res = rep.run(x)
        output_expected = np.array([[49.752754]], dtype=np.float32)
        np.testing.assert_allclose(output_expected, res[0], rtol=1e-05, atol=1e-08)

    def test_run_model_proto(self):
        name = datasets.get_example("logreg_iris.onnx")
        model = load(name)

        rep = backend.prepare(model)
        x = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=np.float32)
        res = rep.run(x)
        output_expected = np.array([0, 0, 0], dtype=np.float32)
        np.testing.assert_allclose(output_expected, res[0], rtol=1e-05, atol=1e-08)
        output_expected = [
            {0: 0.950599730014801, 1: 0.027834169566631317, 2: 0.02156602405011654},
            {
                0: 0.9974970817565918,
                1: 5.6299926654901356e-05,
                2: 0.0024466661270707846,
            },
            {
                0: 0.9997311234474182,
                1: 1.1918064757310276e-07,
                2: 0.00026869276189245284,
            },
        ]

        check_list_of_map_to_float(self, output_expected, res[1])

    def test_run_model_proto_api(self):
        name = datasets.get_example("logreg_iris.onnx")
        model = load(name)

        inputs = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=np.float32)
        outputs = ort_backend.run_model(model, inputs)

        output_expected = np.array([0, 0, 0], dtype=np.float32)
        np.testing.assert_allclose(output_expected, outputs[0], rtol=1e-05, atol=1e-08)
        output_expected = [
            {0: 0.950599730014801, 1: 0.027834169566631317, 2: 0.02156602405011654},
            {
                0: 0.9974970817565918,
                1: 5.6299926654901356e-05,
                2: 0.0024466661270707846,
            },
            {
                0: 0.9997311234474182,
                1: 1.1918064757310276e-07,
                2: 0.00026869276189245284,
            },
        ]

        check_list_of_map_to_float(self, output_expected, outputs[1])


if __name__ == "__main__":
    unittest.main(module=__name__, buffer=True)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import glob
import os
import re
import sys
import zipfile  # Available Python 3.2 or higher

linux_gpu_package_libraries = [
    "libonnxruntime_providers_shared.so",
    "libonnxruntime_providers_cuda.so",
    "libonnxruntime_providers_tensorrt.so",
]
win_gpu_package_libraries = [
    "onnxruntime_providers_shared.lib",
    "onnxruntime_providers_shared.dll",
    "onnxruntime_providers_cuda.lib",
    "onnxruntime_providers_cuda.dll",
    "onnxruntime_providers_tensorrt.lib",
    "onnxruntime_providers_tensorrt.dll",
]
gpu_related_header_files = [
    "cpu_provider_factory.h",
    "onnxruntime_c_api.h",
    "onnxruntime_cxx_api.h",
    "onnxruntime_float16.h",
    "onnxruntime_cxx_inline.h",
]
dmlep_related_header_files = [
    "cpu_provider_factory.h",
    "onnxruntime_c_api.h",
    "onnxruntime_cxx_api.h",
    "onnxruntime_float16.h",
    "onnxruntime_cxx_inline.h",
    "dml_provider_factory.h",
]
training_related_header_files = [
    "onnxruntime_c_api.h",
    "onnxruntime_float16.h",
    "onnxruntime_cxx_api.h",
    "onnxruntime_cxx_inline.h",
    "onnxruntime_training_c_api.h",
    "onnxruntime_training_cxx_api.h",
    "onnxruntime_training_cxx_inline.h",
]


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Validate ONNX Runtime native nuget containing native shared library artifacts spec script",
        usage="",
    )
    # Main arguments
    parser.add_argument("--package_type", required=True, help="Specify nuget, tarball or zip.")
    parser.add_argument("--package_name", required=True, help="Package name to be validated.")
    parser.add_argument(
        "--package_path",
        required=True,
        help="Path containing the package to be validated. Must only contain only one package within this.",
    )
    parser.add_argument(
        "--platforms_supported", required=True, help="Comma separated list (no space). Ex: linux-x64,win-x86,osx-x64"
    )
    parser.add_argument(
        "--verify_nuget_signing",
        help="Flag indicating if Nuget package signing is to be verified. Only accepts 'true' or 'false'",
    )

    return parser.parse_args()


def check_exists(path):
    return os.path.exists(path)


def remove_residual_files(path):
    if check_exists(path):
        os.remove(path)


def is_windows():
    return sys.platform.startswith("win")


def check_if_headers_are_present(header_files, header_folder, file_list_in_package, platform):
    for header in header_files:
        path = header_folder + "/" + header
        print("Checking path: " + path)
        if path not in file_list_in_package:
            print(header + " not found for " + platform)
            raise Exception(header + " not found for " + platform)


def check_if_dlls_are_present(
    package_type,
    is_windows_ai_package,
    is_gpu_package,
    is_dml_package,
    is_training_package,
    platforms_supported,
    zip_file,
    package_path,
    is_gpu_dependent_package=False,  # only used for nuget packages
):
    platforms = platforms_supported.strip().split(",")
    if package_type == "tarball":
        file_list_in_package = []
        for dirpath, _dirnames, filenames in os.walk(package_path):
            file_list_in_package += [os.path.join(dirpath, file) for file in filenames]
    else:
        file_list_in_package = zip_file.namelist()
    print(file_list_in_package)
    # In Nuget GPU package, onnxruntime.dll is in dependent package.
    package_contains_library = not bool(package_type == "nuget" and is_gpu_package)
    # In Nuget GPU package, gpu header files are not in dependent package.
    package_contains_headers = bool(
        (is_gpu_package and package_type != "nuget") or (package_type == "nuget" and not is_gpu_package)
    )
    # In Nuget GPU package, cuda ep and tensorrt ep dlls are in dependent package
    package_contains_cuda_binaries = bool((is_gpu_package and package_type != "nuget") or is_gpu_dependent_package)
    for platform in platforms:
        if platform.startswith("win"):
            native_folder = "_native" if is_windows_ai_package else "native"

            if package_type == "nuget":
                folder = "runtimes/" + platform + "/" + native_folder
                build_dir = "buildTransitive" if is_gpu_dependent_package else "build"
                header_folder = f"{build_dir}/native/include"
            else:  # zip package
                folder = package_path + "/lib"
                header_folder = package_path + "/include"

            # In Nuget GPU package, onnxruntime.dll is in dependent package.
            if package_contains_library:
                path = folder + "/" + "onnxruntime.dll"
                print("Checking path: " + path)
                if path not in file_list_in_package:
                    print("onnxruntime.dll not found for " + platform)
                    raise Exception("onnxruntime.dll not found for " + platform)

            if package_contains_cuda_binaries:
                for dll in win_gpu_package_libraries:
                    path = folder + "/" + dll
                    print("Checking path: " + path)
                    if path not in file_list_in_package:
                        print(dll + " not found for " + platform)
                        raise Exception(dll + " not found for " + platform)

            if package_contains_headers:
                check_if_headers_are_present(gpu_related_header_files, header_folder, file_list_in_package, platform)

            if is_dml_package:
                check_if_headers_are_present(dmlep_related_header_files, header_folder, file_list_in_package, platform)

            if is_training_package:
                check_if_headers_are_present(
                    training_related_header_files, header_folder, file_list_in_package, platform
                )

        elif platform.startswith("linux"):
            if package_type == "nuget":
                folder = "runtimes/" + platform + "/native"
                build_dir = "buildTransitive" if is_gpu_dependent_package else "build"
                header_folder = f"{build_dir}/native/include"
            else:  # tarball package
                folder = package_path + "/lib"
                header_folder = package_path + "/include"

            if package_contains_library:
                path = folder + "/" + "libonnxruntime.so"
                print("Checking path: " + path)
                if path not in file_list_in_package:
                    print("libonnxruntime.so not found for " + platform)
                    raise Exception("libonnxruntime.so not found for " + platform)

            if package_contains_cuda_binaries:
                for so in linux_gpu_package_libraries:
                    path = folder + "/" + so
                    print("Checking path: " + path)
                    if path not in file_list_in_package:
                        print(so + " not found for " + platform)
                        raise Exception(so + " not found for " + platform)

            if package_contains_headers:
                for header in gpu_related_header_files:
                    path = header_folder + "/" + header
                    print("Checking path: " + path)
                    if path not in file_list_in_package:
                        print(header + " not found for " + platform)
                        raise Exception(header + " not found for " + platform)

        elif platform.startswith("osx"):
            path = "runtimes/" + platform + "/native/libonnxruntime.dylib"
            print("Checking path: " + path)
            if path not in file_list_in_package:
                print("libonnxruntime.dylib not found for " + platform)
                raise Exception("libonnxruntime.dylib not found for " + platform)

        else:
            raise Exception("Unsupported platform: " + platform)


def check_if_nuget_is_signed(nuget_path):
    code_sign_summary_file = glob.glob(os.path.join(nuget_path, "*.md"))
    if len(code_sign_summary_file) != 1:
        print("CodeSignSummary files found in path: ")
        print(code_sign_summary_file)
        raise Exception("No CodeSignSummary files / more than one CodeSignSummary files found in the given path.")

    print("CodeSignSummary file: " + code_sign_summary_file[0])

    with open(code_sign_summary_file[0]) as f:
        contents = f.read()
        return "Pass" in contents

    return False


def validate_tarball(args):
    files = glob.glob(os.path.join(args.package_path, args.package_name))
    if len(files) != 1:
        print("packages found in path: ")
        print(files)
        raise Exception("No packages / more than one packages found in the given path.")

    package_name = args.package_name
    if "-gpu-" in package_name.lower():
        is_gpu_package = True
    else:
        is_gpu_package = False

    package_folder = re.search("(.*)[.].*", package_name).group(1)

    print("tar zxvf " + package_name)
    os.system("tar zxvf " + package_name)

    is_windows_ai_package = False
    zip_file = None
    is_dml_package = False
    is_training_package = False
    check_if_dlls_are_present(
        args.package_type,
        is_windows_ai_package,
        is_gpu_package,
        is_dml_package,
        is_training_package,
        args.platforms_supported,
        zip_file,
        package_folder,
    )


def validate_zip(args):
    files = glob.glob(os.path.join(args.package_path, args.package_name))
    if len(files) != 1:
        print("packages found in path: ")
        print(files)
        raise Exception("No packages / more than one packages found in the given path.")

    package_name = args.package_name
    if "-gpu-" in package_name.lower():
        is_gpu_package = True
    else:
        is_gpu_package = False

    package_folder = re.search("(.*)[.].*", package_name).group(1)

    is_windows_ai_package = False
    is_dml_package = False
    is_training_package = False
    zip_file = zipfile.ZipFile(package_name)
    check_if_dlls_are_present(
        args.package_type,
        is_windows_ai_package,
        is_gpu_package,
        is_dml_package,
        is_training_package,
        args.platforms_supported,
        zip_file,
        package_folder,
    )


def validate_nuget(args):
    files = glob.glob(os.path.join(args.package_path, args.package_name))
    nuget_packages_found_in_path = [i for i in files if i.endswith(".nupkg") and "Managed" not in i]
    if len(nuget_packages_found_in_path) != 1:
        print("Nuget packages found in path: ")
        print(nuget_packages_found_in_path)
        raise Exception("No Nuget packages / more than one Nuget packages found in the given path.")
    nuget_file_name = nuget_packages_found_in_path[0]
    full_nuget_path = os.path.join(args.package_path, nuget_file_name)

    is_gpu_package = bool("microsoft.ml.onnxruntime.gpu.1" in args.package_name.lower())
    is_gpu_dependent_package = bool(
        "microsoft.ml.onnxruntime.gpu.windows" in args.package_name.lower()
        or "microsoft.ml.onnxruntime.gpu.linux" in args.package_name.lower()
    )
    if "directml" in nuget_file_name.lower():
        is_dml_package = True
    else:
        is_dml_package = False

    if "Training" in nuget_file_name:
        is_training_package = True
    else:
        is_training_package = False

    exit_code = 0

    nupkg_copy_name = "NugetCopy.nupkg"
    zip_copy_name = "NugetCopy.zip"
    zip_file = None

    # Remove any residual files
    remove_residual_files(nupkg_copy_name)
    remove_residual_files(zip_copy_name)

    # Do all validations here
    try:
        if not is_windows():
            raise Exception("Nuget validation is currently supported only on Windows")

        # Make a copy of the Nuget package
        print("Copying [" + full_nuget_path + "] -> [" + nupkg_copy_name + "], and extracting its contents")
        os.system("copy " + full_nuget_path + " " + nupkg_copy_name)

        # Convert nupkg to zip
        os.rename(nupkg_copy_name, zip_copy_name)
        zip_file = zipfile.ZipFile(zip_copy_name)

        # Check if the relevant dlls are present in the Nuget/Zip
        print("Checking if the Nuget contains relevant dlls")
        is_windows_ai_package = os.path.basename(full_nuget_path).startswith("Microsoft.AI.MachineLearning")
        check_if_dlls_are_present(
            args.package_type,
            is_windows_ai_package,
            is_gpu_package,
            is_dml_package,
            is_training_package,
            args.platforms_supported,
            zip_file,
            None,
            is_gpu_dependent_package,
        )

        verify_nuget_signing = args.verify_nuget_signing.lower()
        # Check if the Nuget has been signed
        if verify_nuget_signing != "true" and verify_nuget_signing != "false":
            raise Exception("Parameter verify_nuget_signing accepts only true or false as an argument")

        if verify_nuget_signing == "true":
            print("Verifying if Nuget has been signed")
            if not check_if_nuget_is_signed(args.package_path):
                print("Nuget signing verification failed")
                raise Exception("Nuget signing verification failed")

    except Exception as e:
        print(e)
        exit_code = 1

    finally:
        print("Cleaning up after Nuget validation")

        if zip_file is not None:
            zip_file.close()

        if check_exists(zip_copy_name):
            os.remove(zip_copy_name)

        if exit_code == 0:
            print("Nuget validation was successful")
        else:
            raise Exception("Nuget validation was unsuccessful")


def main():
    args = parse_arguments()

    if args.package_type == "nuget":
        validate_nuget(args)
    elif args.package_type == "tarball":
        validate_tarball(args)
    elif args.package_type == "zip":
        validate_zip(args)
    else:
        print(f"Package type {args.package_type} is not supported")


if __name__ == "__main__":
    sys.exit(main())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import re
from dataclasses import dataclass
from itertools import product

import kernel_explorer as ke
import numpy as np
import pytest
from utils import dtype_to_bytes, dtype_to_suffix, standardization


def get_sd_sizes():
    batch_sizes = [1, 2]
    height = [8, 16, 32]
    num_channels = [320, 640, 1280, 1920, 2560]

    num_groups = [32]
    return product(batch_sizes, height, num_channels, num_groups)


def dtype_to_funcs(dtype):
    type_map = {
        "float16": list(filter(lambda x: re.match("GroupNormNHWC.*_half", x), dir(ke))),
        "float32": list(filter(lambda x: re.match("GroupNormNHWC.*_float", x), dir(ke))),
    }
    return type_map[dtype]


def sigmoid_function(x):
    return 1.0 / (1.0 + np.exp(-x))


def group_norm(input_x, skip_x, bias_x, gamma, beta, num_groups, epsilon, with_silu, has_skip):
    add_output = None
    if has_skip:
        input_x = input_x + skip_x + bias_x
        add_output = input_x
    n, h, w, c = input_x.shape
    input_x = input_x.transpose([0, 3, 1, 2])
    assert c % num_groups == 0
    x = input_x.reshape((n, num_groups, -1))
    x = standardization(x, -1, epsilon)
    x = x.reshape((n, c, h, w))
    x = x.transpose([0, 2, 3, 1])
    x = x * gamma + beta

    if with_silu:
        x = x * sigmoid_function(x)
    return x, add_output


def run_group_norm(
    batch_size: int, height: int, num_channels: int, num_groups: int, dtype: str, silu: bool, has_skip: bool, func
):
    np.random.seed(0)
    width = height
    input_x = np.random.rand(batch_size, height, width, num_channels).astype(np.float32)
    gamma = np.random.rand(num_channels).astype(np.float32)
    beta = np.random.rand(num_channels).astype(np.float32)
    # the size of workspace is defined in onnxruntime/contrib_ops/cuda/diffusion/group_norm_impl.h L18
    workspace = np.random.rand((np.dtype(np.float32).itemsize * 2) * batch_size * num_groups).astype(np.float32)
    epsilon = 1e-05
    output_y = np.random.rand(batch_size, height, width, num_channels).astype(dtype)

    skip_x = (
        np.random.rand(batch_size, height, width, num_channels).astype(np.float32)
        if has_skip
        else np.empty((0), dtype=dtype)
    )
    bias_x = np.random.rand(num_channels).astype(np.float32) if has_skip else np.empty((0), dtype=dtype)
    add_output = (
        np.random.rand(batch_size, height, width, num_channels).astype(dtype)
        if has_skip
        else np.empty((0), dtype=dtype)
    )
    use_silu = silu
    broadcast_skip = False
    if has_skip:
        skip_x_shape = skip_x.shape
        b2 = len(skip_x_shape) == 2 and skip_x_shape[0] == batch_size and skip_x_shape[1] == num_channels
        b4 = (
            len(skip_x_shape) == 4
            and skip_x_shape[0] == batch_size
            and skip_x_shape[1] == 1
            and skip_x_shape[2] == 1
            and skip_x_shape[3] == num_channels
        )
        if b2 or b4:
            broadcast_skip = True
    channels_per_block = 0  # Compute in params initialization

    input_d = ke.DeviceArray(input_x.astype(dtype))
    skip_d = ke.DeviceArray(skip_x.astype(dtype))
    bias_d = ke.DeviceArray(bias_x.astype(dtype))
    gamma_d = ke.DeviceArray(gamma)
    beta_d = ke.DeviceArray(beta)
    workspace_d = ke.DeviceArray(workspace)
    y_d = ke.DeviceArray(output_y)
    y_add_d = ke.DeviceArray(add_output)
    f = getattr(ke, func)

    my_op = f(
        y_d,
        y_add_d,
        input_d,
        skip_d,
        bias_d,
        gamma_d,
        beta_d,
        workspace_d,
        epsilon,
        batch_size,
        num_channels,
        height,
        width,
        num_groups,
        use_silu,
        broadcast_skip,
        channels_per_block,
    )
    y_ref, y_add_d_ref = group_norm(input_x, skip_x, bias_x, gamma, beta, num_groups, epsilon, use_silu, has_skip)
    y_ref = y_ref.astype(dtype)

    for impl in my_op.ListOps():
        if not my_op.SelectOp(impl):
            continue

        my_op.Run()

        y_d.UpdateHostNumpyArray()

        np.testing.assert_allclose(y_ref, output_y, atol=1e-02)
        if has_skip:
            y_add_d_ref = y_add_d_ref.astype(dtype)
            y_add_d.UpdateHostNumpyArray()
            np.testing.assert_allclose(y_add_d_ref, add_output, atol=1e-02)


dtypes = ["float32", "float16"]


@pytest.mark.parametrize("sd_sizes", get_sd_sizes())
@pytest.mark.parametrize("dtype", dtypes)
@pytest.mark.parametrize("silu", [True])
@pytest.mark.parametrize("has_skip", [True, False])
def test_group_norm(sd_sizes, dtype, silu, has_skip):
    for func in dtype_to_funcs(dtype):
        run_group_norm(*sd_sizes, dtype, silu, has_skip, func)


@pytest.mark.parametrize("sd_sizes", get_sd_sizes())
@pytest.mark.parametrize("dtype", dtypes)
@pytest.mark.parametrize("silu", [True])
@pytest.mark.parametrize("has_skip", [False])
def test_group_norm_ck(sd_sizes, dtype, silu, has_skip):
    silu_suffix = "Silu" if silu else "Pass"
    ck_f_name = "CKGroupNormNHWC" + silu_suffix + "_" + dtype_to_suffix(dtype)
    run_group_norm(*sd_sizes, dtype, silu, has_skip, ck_f_name)


@dataclass
class GroupNormNHWCMetric(ke.BandwidthMetric):
    batch_size: int
    height: int
    width: int
    num_channels: int
    groups: int

    def report(self):
        common = (
            f"{self.dtype:<4} batch={self.batch_size:<4} height={self.height:<4} width={self.width:<4}"
            f"num_channels={self.num_channels:<6} groups={self.groups:<4} {self.name}"
        )
        if self.duration > 0:
            return f"{self.duration:>6.2f} us, {self.gbps:>5.2f} GB/s  " + common
        return "not supported          " + common


@ke.dispatchable(pattern_arg=8)
def profile_group_norm_func(
    batch_size: int,
    height: int,
    width: int,
    num_channels: int,
    num_groups: int,
    dtype: str,
    silu: bool,
    has_skip: bool,
    func,
):
    np.random.seed(0)
    input_x = np.random.rand(batch_size, height, width, num_channels).astype(dtype)
    gamma = np.random.rand(num_channels).astype(np.float32)
    beta = np.random.rand(num_channels).astype(np.float32)
    workspace = np.random.rand(np.dtype(np.float32).itemsize * 2 * batch_size * num_groups).astype(np.float32)
    epsilon = 0.05
    output_y = np.random.rand(batch_size, height, width, num_channels).astype(dtype)

    skip_x = (
        np.random.rand(batch_size, height, width, num_channels).astype(dtype)
        if has_skip
        else np.empty((0), dtype=dtype)
    )
    bias_x = np.random.rand(num_channels).astype(dtype) if has_skip else np.empty((0), dtype=dtype)
    add_output = (
        np.random.rand(batch_size, height, width, num_channels).astype(dtype)
        if has_skip
        else np.empty((0), dtype=dtype)
    )
    use_silu = silu
    broadcast_skip = False
    channels_per_block = 0  # Compute in params initialization

    input_d = ke.DeviceArray(input_x)
    skip_d = ke.DeviceArray(skip_x)
    bias_d = ke.DeviceArray(bias_x)
    gamma_d = ke.DeviceArray(gamma)
    beta_d = ke.DeviceArray(beta)
    workspace_d = ke.DeviceArray(workspace)
    y_d = ke.DeviceArray(output_y)
    y_add_d = ke.DeviceArray(add_output)
    f = getattr(ke, func)

    my_op = f(
        y_d,
        y_add_d,
        input_d,
        skip_d,
        bias_d,
        gamma_d,
        beta_d,
        workspace_d,
        epsilon,
        batch_size,
        num_channels,
        height,
        width,
        num_groups,
        use_silu,
        broadcast_skip,
        channels_per_block,
    )
    for impl in my_op.ListOps():
        duration_ms = -1
        if my_op.SelectOp(impl):
            duration_ms = my_op.Profile()
        total_bytes = (input_x.size * 2 + gamma.size * 2) * dtype_to_bytes(dtype)

        ke.report(
            GroupNormNHWCMetric(
                impl, dtype, duration_ms, total_bytes, batch_size, height, width, num_channels, num_groups
            )
        )


@ke.dispatchable
def profile_with_args(batch_size, height, width, num_channels, num_groups, dtype, silu=True, has_skip=True):
    with ke.benchmark():
        for func in dtype_to_funcs(dtype):
            profile_group_norm_func(batch_size, height, width, num_channels, num_groups, dtype, silu, has_skip, func)
        # ck function
        silu_suffix = "Silu" if silu else "Pass"
        ck_f_name = "CKGroupNormNHWC" + silu_suffix + "_" + dtype_to_suffix(dtype)
        profile_group_norm_func(batch_size, height, width, num_channels, num_groups, dtype, silu, has_skip, ck_f_name)


sd_profile_sizes = [
    (2, 64, 64, 320, 32),
    (2, 32, 32, 640, 32),
    (2, 16, 16, 1280, 32),
    (2, 64, 64, 640, 32),
    (2, 16, 16, 2560, 32),
    (2, 32, 32, 1280, 32),
    (2, 32, 32, 1920, 32),
    (2, 8, 8, 1280, 32),
    (2, 64, 64, 960, 32),
    (2, 32, 32, 960, 32),
    (2, 32, 32, 320, 32),
    (2, 16, 16, 640, 32),
    (2, 16, 16, 1920, 32),
    (2, 8, 8, 2560, 32),
]


def profile():
    for dtype in dtypes:
        for sd_size in sd_profile_sizes:
            profile_with_args(*sd_size, dtype)
            print()


if __name__ == "__main__":
    parser = ke.get_argument_parser()
    group = parser.add_argument_group()
    group.add_argument("batch_size", type=int)
    group.add_argument("height", type=int)
    group.add_argument("width", type=int)
    group.add_argument("num_channels", type=int)
    group.add_argument("num_groups", type=int)
    group.add_argument("dtype", choices=dtypes)
    group.add_argument("--silu", action="store_true")
    group.add_argument("--has_skip", action="store_true")

    if not ke.has_args():
        profile()
    else:
        args = parser.parse_args()
        args.dispatch(
            args.batch_size,
            args.height,
            args.width,
            args.num_channels,
            args.num_groups,
            args.dtype,
            args.silu,
            args.has_skip,
        )



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import math
import unittest

import numpy as np
import onnx

import onnxruntime
from onnxruntime.quantization.execution_providers.qnn.fusion_lpnorm import FusionLpNormalization
from onnxruntime.quantization.fusions import FusionGelu, FusionLayerNormalization
from onnxruntime.quantization.onnx_model import ONNXModel


class TestFusions(unittest.TestCase):
    def check_fused_model_correctness(self, orig_model, fused_model, inputs, rtol=1e-7, atol=0):
        """
        Checks that the output of the fused model matches the output of the original model.
        """
        orig_session = onnxruntime.InferenceSession(orig_model.SerializeToString(), providers=["CPUExecutionProvider"])
        orig_results = orig_session.run(None, inputs)

        fused_session = onnxruntime.InferenceSession(
            fused_model.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        fused_results = fused_session.run([], inputs)

        self.assertEqual(len(orig_results), len(fused_results), "Number of outputs for fused model differs")
        for idx, expected_output in enumerate(orig_results):
            actual_output = fused_results[idx]
            np.testing.assert_allclose(
                expected_output,
                actual_output,
                rtol=rtol,
                atol=atol,
                err_msg=f"Fused model output {idx} differs",
            )

    def build_erf_sequence_1_model(self, shape):
        """
        Erf sequence that fuses into Gelu:
           +-------Mul(0.5)---------------------+
           |                                    |
           |                                    v
        [root] --> Div -----> Erf  --> Add --> Mul -->
                  (B=1.4142...)       (1)

        This method builds 2 of these Erf sequences:

        [root] -> ERF_SEQUENCE1 -> ERF_SEQUENCE2 -> output
        """
        root_inp = onnx.helper.make_tensor_value_info("root", onnx.TensorProto.FLOAT, shape)
        output = onnx.helper.make_tensor_value_info("output", onnx.TensorProto.FLOAT, shape)
        one_const = onnx.numpy_helper.from_array(np.array(1.0, dtype=np.float32), "one_const")
        half_const = onnx.numpy_helper.from_array(np.array(0.5, dtype=np.float32), "half_const")
        root2_const = onnx.numpy_helper.from_array(np.array(math.sqrt(2.0), dtype=np.float32), "root2_const")

        # First Erf sequence
        mul0_node = onnx.helper.make_node("Mul", ["root", "half_const"], ["mul0_out"])
        div_node = onnx.helper.make_node("Div", ["root", "root2_const"], ["div_out"])
        erf_node = onnx.helper.make_node("Erf", ["div_out"], ["erf_out"])
        add_node = onnx.helper.make_node("Add", ["erf_out", "one_const"], ["add_out"])
        mul1_node = onnx.helper.make_node("Mul", ["add_out", "mul0_out"], ["seq1_output"])

        # Second Erf sequence
        mul0_node_dup = onnx.helper.make_node("Mul", ["seq1_output", "half_const"], ["mul0_out_dup"])
        div_node_dup = onnx.helper.make_node("Div", ["seq1_output", "root2_const"], ["div_out_dup"])
        erf_node_dup = onnx.helper.make_node("Erf", ["div_out_dup"], ["erf_out_dup"])
        add_node_dup = onnx.helper.make_node("Add", ["erf_out_dup", "one_const"], ["add_out_dup"])
        mul1_node_dup = onnx.helper.make_node("Mul", ["add_out_dup", "mul0_out_dup"], ["output"])

        graph = onnx.helper.make_graph(
            [
                mul0_node,
                div_node,
                erf_node,
                add_node,
                mul1_node,
                mul0_node_dup,
                div_node_dup,
                erf_node_dup,
                add_node_dup,
                mul1_node_dup,
            ],
            "two_erf_sequences",
            [root_inp],
            [output],
            initializer=[one_const, half_const, root2_const],
        )
        opset_imports = [
            onnx.helper.make_opsetid("", 18),
            onnx.helper.make_opsetid("com.microsoft", 1),
        ]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        return ONNXModel(model)

    def build_erf_sequence_2_model(self, shape):
        """
           +------------------------------------+
           |                                    |
           |                                    v
        [root] --> Div -----> Erf  --> Add --> Mul -->Mul -->
                  (B=1.4142...)       (1)            (0.5)

        """
        root_inp = onnx.helper.make_tensor_value_info("root", onnx.TensorProto.FLOAT, shape)
        output = onnx.helper.make_tensor_value_info("output", onnx.TensorProto.FLOAT, shape)
        one_const = onnx.numpy_helper.from_array(np.array(1.0, dtype=np.float32), "one_const")
        half_const = onnx.numpy_helper.from_array(np.array(0.5, dtype=np.float32), "half_const")
        root2_const = onnx.numpy_helper.from_array(np.array(math.sqrt(2.0), dtype=np.float32), "root2_const")

        div_node = onnx.helper.make_node("Div", ["root", "root2_const"], ["div_out"])
        erf_node = onnx.helper.make_node("Erf", ["div_out"], ["erf_out"])
        add_node = onnx.helper.make_node("Add", ["erf_out", "one_const"], ["add_out"])
        mul0_node = onnx.helper.make_node("Mul", ["add_out", "root"], ["mul0_out"])
        mul1_node = onnx.helper.make_node("Mul", ["mul0_out", "half_const"], ["output"])

        graph = onnx.helper.make_graph(
            [div_node, erf_node, add_node, mul0_node, mul1_node],
            "erf_sequence_2",
            [root_inp],
            [output],
            initializer=[one_const, half_const, root2_const],
        )
        opset_imports = [
            onnx.helper.make_opsetid("", 18),
            onnx.helper.make_opsetid("com.microsoft", 1),
        ]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        return ONNXModel(model)

    def build_erf_sequence_3_model(self, shape):
        """
           +------------------------------------------+
           |                                          |
           |                                          v
        [root] --> Div -----> Erf  --> Add --> Mul -->Mul
                  (B=1.4142...)       (A=1)   (A=0.5)

        """
        root_inp = onnx.helper.make_tensor_value_info("root", onnx.TensorProto.FLOAT, shape)
        output = onnx.helper.make_tensor_value_info("output", onnx.TensorProto.FLOAT, shape)
        one_const = onnx.numpy_helper.from_array(np.array(1.0, dtype=np.float32), "one_const")
        half_const = onnx.numpy_helper.from_array(np.array(0.5, dtype=np.float32), "half_const")
        root2_const = onnx.numpy_helper.from_array(np.array(math.sqrt(2.0), dtype=np.float32), "root2_const")

        div_node = onnx.helper.make_node("Div", ["root", "root2_const"], ["div_out"])
        erf_node = onnx.helper.make_node("Erf", ["div_out"], ["erf_out"])
        add_node = onnx.helper.make_node("Add", ["erf_out", "one_const"], ["add_out"])
        mul0_node = onnx.helper.make_node("Mul", ["add_out", "half_const"], ["mul0_out"])
        mul1_node = onnx.helper.make_node("Mul", ["mul0_out", "root"], ["output"])

        graph = onnx.helper.make_graph(
            [div_node, erf_node, add_node, mul0_node, mul1_node],
            "erf_sequence_3",
            [root_inp],
            [output],
            initializer=[one_const, half_const, root2_const],
        )
        opset_imports = [
            onnx.helper.make_opsetid("", 18),
            onnx.helper.make_opsetid("com.microsoft", 1),
        ]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        return ONNXModel(model)

    def build_erf_sequence_4_model(self, shape):
        """
           +----------------------------------------------+
           |                                              |
           |                                              v
        [root] --> Mul -----> Erf    -->   Add --> Mul -->Mul
                   (A=0.7071067690849304)  (B=1)  (B=0.5)

        """
        root_inp = onnx.helper.make_tensor_value_info("root", onnx.TensorProto.FLOAT, shape)
        output = onnx.helper.make_tensor_value_info("output", onnx.TensorProto.FLOAT, shape)
        one_const = onnx.numpy_helper.from_array(np.array(1.0, dtype=np.float32), "one_const")
        half_const = onnx.numpy_helper.from_array(np.array(0.5, dtype=np.float32), "half_const")
        frac_const = onnx.numpy_helper.from_array(np.array(0.7071067690849304, dtype=np.float32), "frac_const")

        mul0_node = onnx.helper.make_node("Mul", ["root", "frac_const"], ["mul0_out"])
        erf_node = onnx.helper.make_node("Erf", ["mul0_out"], ["erf_out"])
        add_node = onnx.helper.make_node("Add", ["erf_out", "one_const"], ["add_out"])
        mul1_node = onnx.helper.make_node("Mul", ["add_out", "half_const"], ["mul1_out"])
        mul2_node = onnx.helper.make_node("Mul", ["mul1_out", "root"], ["output"])

        graph = onnx.helper.make_graph(
            [mul0_node, erf_node, add_node, mul1_node, mul2_node],
            "erf_sequence_4",
            [root_inp],
            [output],
            initializer=[one_const, half_const, frac_const],
        )
        opset_imports = [
            onnx.helper.make_opsetid("", 18),
            onnx.helper.make_opsetid("com.microsoft", 1),
        ]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        return ONNXModel(model)

    def build_reduce_mean_sequence_model(self, shape, scale_val, bias_val, axis=-1):
        """
            +----------------------+
            |                      |
            |                      v
        [Root] --> ReduceMean -->  Sub  --> Pow --> ReduceMean --> Add --> Sqrt --> Div --> Mul --> Add
                   (axis=2 or -1)  |      (Y=2)   (axis=2 or -1)  (E-6 or E-12 or 0) ^       ^       ^
                                   |                                                 |       |       |
                                   +-------------------------------------------------+    [Scale]  [Bias]
        """
        root_inp = onnx.helper.make_tensor_value_info("root", onnx.TensorProto.FLOAT, shape)
        output = onnx.helper.make_tensor_value_info("output", onnx.TensorProto.FLOAT, shape)
        scale_const = onnx.numpy_helper.from_array(np.array(scale_val, dtype=np.float32), "scale_const")
        bias_const = onnx.numpy_helper.from_array(np.array(bias_val, dtype=np.float32), "bias_const")
        axes_const = onnx.numpy_helper.from_array(np.array([axis], dtype=np.int64), "axes_const")
        two_const = onnx.numpy_helper.from_array(np.array(2.0, dtype=np.float32), "two_const")
        eps_const = onnx.numpy_helper.from_array(np.array(1.0e-8, dtype=np.float32), "eps_const")

        rm0_node = onnx.helper.make_node("ReduceMean", ["root", "axes_const"], ["rm0_out"])
        sub_node = onnx.helper.make_node("Sub", ["root", "rm0_out"], ["sub_out"])
        pow_node = onnx.helper.make_node("Pow", ["sub_out", "two_const"], ["pow_out"])
        rm1_node = onnx.helper.make_node("ReduceMean", ["pow_out", "axes_const"], ["rm1_out"])
        add0_node = onnx.helper.make_node("Add", ["rm1_out", "eps_const"], ["add0_out"])
        sqrt_node = onnx.helper.make_node("Sqrt", ["add0_out"], ["sqrt_out"])
        div_node = onnx.helper.make_node("Div", ["sub_out", "sqrt_out"], ["div_out"])
        mul_node = onnx.helper.make_node("Mul", ["div_out", "scale_const"], ["mul_out"])
        add1_node = onnx.helper.make_node("Add", ["mul_out", "bias_const"], ["output"])

        graph = onnx.helper.make_graph(
            [rm0_node, sub_node, pow_node, rm1_node, add0_node, sqrt_node, div_node, mul_node, add1_node],
            "reduce_mean_sequence",
            [root_inp],
            [output],
            initializer=[scale_const, bias_const, axes_const, two_const, eps_const],
        )
        opset_imports = [
            onnx.helper.make_opsetid("", 18),
        ]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        return ONNXModel(model)

    def build_reduce_l2_sequence_model(self, shape, epsilon_val, axis=-1):
        """
        [root] --> ReduceL2 -----> Clip  --> Expand ----> Div -->
           |      (axis=-1)    (min=epsilon) (shape=root)  ^
           |   (keepdims=True)                             |
           |                                               |
           +-----------------------------------------------+
        """
        root_inp = onnx.helper.make_tensor_value_info("root", onnx.TensorProto.FLOAT, shape)
        output = onnx.helper.make_tensor_value_info("output", onnx.TensorProto.FLOAT, shape)
        axes_const = onnx.numpy_helper.from_array(np.array([axis], dtype=np.int64), "axes_const")
        eps_const = onnx.numpy_helper.from_array(np.array(epsilon_val, dtype=np.float32), "eps_const")
        shape_const = onnx.numpy_helper.from_array(np.array(list(shape), dtype=np.int64), "shape_const")

        rl2_node = onnx.helper.make_node("ReduceL2", ["root", "axes_const"], ["rl2_out"], keepdims=1)
        clip_node = onnx.helper.make_node("Clip", ["rl2_out", "eps_const"], ["clip_out"])
        expand_node = onnx.helper.make_node("Expand", ["clip_out", "shape_const"], ["expand_out"])
        div_node = onnx.helper.make_node("Div", ["root", "expand_out"], ["output"])

        graph = onnx.helper.make_graph(
            [rl2_node, clip_node, expand_node, div_node],
            "reducel2_sequence",
            [root_inp],
            [output],
            initializer=[axes_const, eps_const, shape_const],
        )
        opset_imports = [
            onnx.helper.make_opsetid("", 18),
        ]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        return ONNXModel(model)

    def test_fuse_erf_to_gelu_1(self):
        shape = (1, 2, 3)
        model = self.build_erf_sequence_1_model(shape)
        orig_model = onnx.ModelProto()
        orig_model.CopyFrom(model.model)

        # Check that fusion simplified model to 2 Gelu nodes.
        modified = FusionGelu(model).apply()
        self.assertTrue(modified)
        self.assertEqual(len(model.model.graph.node), 2)

        gelu_node_0 = model.model.graph.node[0]
        gelu_node_1 = model.model.graph.node[1]
        self.assertEqual(gelu_node_0.op_type, "Gelu")
        self.assertEqual(gelu_node_1.op_type, "Gelu")

        self.assertTrue(gelu_node_0.name)
        self.assertTrue(gelu_node_1.name)
        self.assertNotEqual(gelu_node_0.name, gelu_node_1.name)  # Generated names should not be equal

        # Check that fusion is equivalent to original Erf model.
        inputs = {"root": np.ones(shape, dtype=np.float32)}
        self.check_fused_model_correctness(orig_model, model.model, inputs)

    def test_fuse_erf_to_gelu_2(self):
        shape = (1, 2, 3)
        model = self.build_erf_sequence_2_model(shape)
        orig_model = onnx.ModelProto()
        orig_model.CopyFrom(model.model)

        # Check that fusion simplified model to 1 Gelu node.
        modified = FusionGelu(model).apply()
        self.assertTrue(modified)
        self.assertEqual(len(model.model.graph.node), 1)

        gelu_node = model.model.graph.node[0]
        self.assertEqual(gelu_node.op_type, "Gelu")
        self.assertTrue(gelu_node.name)

        # Check that fusion is equivalent to original Erf model.
        inputs = {"root": np.ones(shape, dtype=np.float32)}
        self.check_fused_model_correctness(orig_model, model.model, inputs)

    def test_fuse_erf_to_gelu_3(self):
        shape = (1, 2, 3)
        model = self.build_erf_sequence_3_model(shape)
        orig_model = onnx.ModelProto()
        orig_model.CopyFrom(model.model)

        # Check that fusion simplified model to 1 Gelu node.
        modified = FusionGelu(model).apply()
        self.assertTrue(modified)
        self.assertEqual(len(model.model.graph.node), 1)

        gelu_node = model.model.graph.node[0]
        self.assertEqual(gelu_node.op_type, "Gelu")
        self.assertTrue(gelu_node.name)

        # Check that fusion is equivalent to original Erf model.
        inputs = {"root": np.ones(shape, dtype=np.float32)}
        self.check_fused_model_correctness(orig_model, model.model, inputs)

    def test_fuse_erf_to_gelu_4(self):
        shape = (1, 2, 3)
        model = self.build_erf_sequence_4_model(shape)
        orig_model = onnx.ModelProto()
        orig_model.CopyFrom(model.model)

        # Check that fusion simplified model to 1 Gelu node.
        modified = FusionGelu(model).apply()
        self.assertTrue(modified)
        self.assertEqual(len(model.model.graph.node), 1)

        gelu_node = model.model.graph.node[0]
        self.assertEqual(gelu_node.op_type, "Gelu")
        self.assertTrue(gelu_node.name)

        # Check that fusion is equivalent to original Erf model.
        inputs = {"root": np.ones(shape, dtype=np.float32)}
        self.check_fused_model_correctness(orig_model, model.model, inputs)

    def test_fuse_reduce_l2_to_lpnorm(self):
        shape = (1, 2, 3)
        model = self.build_reduce_l2_sequence_model(shape, 1e-12, axis=-1)
        orig_model = onnx.ModelProto()
        orig_model.CopyFrom(model.model)

        # Check that fusion simplified model to 1 LpNormalization node.
        modified = FusionLpNormalization(model).apply()
        self.assertTrue(modified)
        self.assertEqual(len(model.model.graph.node), 1)

        lpnorm_node = model.model.graph.node[0]
        self.assertEqual(lpnorm_node.op_type, "LpNormalization")
        self.assertTrue(lpnorm_node.name)

        # LpNorm's p attribute should be set to 2
        p_attr = next(attr for attr in lpnorm_node.attribute if attr.name == "p")
        self.assertEqual(p_attr.i, 2)

    def test_fuse_reduce_mean_to_layer_norm(self):
        shape = (1, 2, 3)
        model = self.build_reduce_mean_sequence_model(shape, [2.0, 2.0, 2.0], [1.0, 1.0, 1.0], axis=-1)
        orig_model = onnx.ModelProto()
        orig_model.CopyFrom(model.model)

        # Check that fusion simplified model to 1 LayerNormalization node.
        modified = FusionLayerNormalization(model).apply()
        self.assertTrue(modified)
        self.assertEqual(len(model.model.graph.node), 1)

        layer_norm_node = model.model.graph.node[0]
        self.assertEqual(layer_norm_node.op_type, "LayerNormalization")
        self.assertTrue(layer_norm_node.name)

        # Check that fused model is equivalent to original model.
        inputs = {"root": np.ones(shape, dtype=np.float32)}
        self.check_fused_model_correctness(orig_model, model.model, inputs)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os
import shutil
import sys


# Note: This script is mainly used for sanity checking/validating the files in the .a library equal to the .o files
# in the source dir to handle the case of source files having duplicate names under different subdirectories for
# each onnxruntime library. (Only applicable when doing a Mac Catalyst build.)
def main():
    source_dir = sys.argv[1]
    dest_dir = sys.argv[2]
    files_from_static_lib = sys.argv[3]
    files_from_source_dir = []
    for subdir, _, files in os.walk(source_dir):
        for file_name in files:
            if file_name.endswith(".o"):
                files_from_source_dir.append(file_name.strip())
                dest_name_without_extension, _ = os.path.splitext(file_name)
                counter = 0

                dest_file = f"{dest_name_without_extension}.o"
                while os.path.exists(os.path.join(dest_dir, dest_file)):
                    print("Duplicate file name from source: " + os.path.join(source_dir, subdir, file_name))
                    counter += 1
                    dest_file = f"{dest_name_without_extension}_{counter}.o"
                    print("Renamed file name in destination: " + os.path.join(dest_dir, dest_file))

                destination_path = os.path.join(dest_dir, dest_file)
                source_file = os.path.join(source_dir, subdir, file_name)
                shutil.copy(source_file, destination_path)

    # Sanity check to ensure the number of .o object from the original cmake source directory matches with the number
    # of .o files extracted from each .a onnxruntime library
    file_lists_from_static_lib = []
    with open(files_from_static_lib) as file:
        filenames = file.readlines()
    for filename in filenames:
        file_lists_from_static_lib.append(filename.strip())

    sorted_list1 = sorted(file_lists_from_static_lib)
    sorted_list2 = sorted(files_from_source_dir)

    if len(sorted_list1) != len(sorted_list2):
        print(
            "Caught a mismatch in the number of .o object files from the original cmake source directory: ",
            len(sorted_list1),
            "the number of .o files extracted from the static onnxruntime lib: ",
            len(sorted_list2),
            "for: ",
            os.path.basename(source_dir),
        )

    if sorted_list1 == sorted_list2:
        print(
            "Sanity check passed: object files from original source directory matches with files extracted "
            "from static library for: ",
            os.path.basename(source_dir),
        )
    else:
        print(
            "Error: Mismatch between object files from original source directory "
            "and the .o files extracted from static library for: ",
            os.path.basename(source_dir),
        )


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import os

import onnx
import torch
from transformers.modeling_utils import Conv1D

logger = logging.getLogger(__name__)


def _conv1d_to_linear(module):
    in_size, out_size = module.weight.shape
    linear = torch.nn.Linear(in_size, out_size)
    linear.weight.data = module.weight.data.T.contiguous()
    linear.bias.data = module.bias.data
    return linear


def conv1d_to_linear(model):
    """in-place
    This is for Dynamic Quantization, as Conv1D is not recognized by PyTorch, convert it to nn.Linear
    """
    logger.debug("replace Conv1D with Linear")
    for name in list(model._modules):
        module = model._modules[name]
        if isinstance(module, Conv1D):
            linear = _conv1d_to_linear(module)
            model._modules[name] = linear
        else:
            conv1d_to_linear(module)


def _get_size_of_pytorch_model(model):
    torch.save(model.state_dict(), "temp.p")
    size = os.path.getsize("temp.p") / (1024 * 1024)
    os.remove("temp.p")
    return size


class QuantizeHelper:
    @staticmethod
    def quantize_torch_model(model, dtype=torch.qint8):
        """
        Usage: model = quantize_model(model)

        TODO: mix of in-place and return, but results are different
        """
        conv1d_to_linear(model)
        quantized_model = torch.quantization.quantize_dynamic(model, {torch.nn.Linear}, dtype=dtype)
        logger.info(f"Size of full precision Torch model(MB):{_get_size_of_pytorch_model(model)}")
        logger.info(f"Size of quantized Torch model(MB):{_get_size_of_pytorch_model(quantized_model)}")
        return quantized_model

    @staticmethod
    def quantize_onnx_model(onnx_model_path, quantized_model_path, use_external_data_format=False):
        from pathlib import Path

        from onnxruntime.quantization import quantize_dynamic

        Path(quantized_model_path).parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Size of full precision ONNX model(MB):{os.path.getsize(onnx_model_path) / (1024 * 1024)}")
        quantize_dynamic(
            onnx_model_path,
            quantized_model_path,
            use_external_data_format=use_external_data_format,
            extra_options={"DefaultTensorType": onnx.TensorProto.FLOAT},
        )
        logger.info(f"quantized model saved to:{quantized_model_path}")
        # TODO: inlcude external data in total model size.
        logger.info(f"Size of quantized ONNX model(MB):{os.path.getsize(quantized_model_path) / (1024 * 1024)}")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


import onnx


def get_output_from_output_name(onnx_model: onnx.ModelProto, output_name: str) -> onnx.ValueInfoProto:
    """Returns the graph output for the given output name"""

    for go in onnx_model.graph.output:
        if go.name == output_name:
            return go

    raise LookupError(f"The provided output name {output_name} is not a graph output.")


def get_input_from_input_name(onnx_model: onnx.ModelProto, input_name: str) -> onnx.ValueInfoProto:
    """Returns the graph input for the given input name"""

    for gi in onnx_model.graph.input:
        if gi.name == input_name:
            return gi

    raise LookupError(f"The provided output name {input_name} is not a graph input.")


_GRAPH_TOKEN = 0


def _get_token() -> int:
    """Return a token that is one more than the previous token retrieved by calling this function."""

    global _GRAPH_TOKEN  # pylint: disable=global-statement  # noqa: PLW0603
    _GRAPH_TOKEN += 1
    return _GRAPH_TOKEN


def generate_graph_name(token: str) -> str:
    """Return a string that can be used in the graph as a graph attribute name."""

    return f"onnx::{token}::{_get_token()}"


def register_graph_outputs(model: onnx.ModelProto, output_names: list[str] | str) -> None:
    """Register the given output names as graph outputs.

    The graph outputs shape information is extracted from the graph value_infos and
    existing graph outputs. The graph output can only be added to the
    graph for those outputs whose value info is known. If value info
    is not known, an error will be raised.
    """

    if isinstance(output_names, str):
        output_names = [output_names]

    name_value_info_mapping = {value_info.name: value_info for value_info in model.graph.value_info}
    name_graph_output_mapping = {output.name: output for output in model.graph.output}

    # collect all new graph outputs (i.e. graph outputs that are not
    # already graph outputs)
    graph_outputs = []
    for output_name in output_names:
        if output_name in name_graph_output_mapping:
            graph_outputs.append(name_graph_output_mapping[output_name])
        elif output_name in name_value_info_mapping:
            graph_outputs.append(name_value_info_mapping[output_name])
        else:
            raise LookupError(f"The provided name {output_name} is not a graph value info or a graph output.")

    del model.graph.output[:]

    model.graph.output.extend(graph_outputs)


def node_arg_exists(model: onnx.ModelProto, node_arg_name: str) -> bool:
    """Returns True if the given node_arg_name exists in the model graph."""

    for node in model.graph.node:
        if node_arg_name in node.input:
            return True

        if node_arg_name in node.output:
            return True

    return False



import argparse

import onnx

parser = argparse.ArgumentParser(description="ONNX file analyzer for performance investigation.")
parser.add_argument("onnx_file", type=str, help="ONNX file to analyze")
args = parser.parse_args()


def process_file(onnx_file):
    model = onnx.load(onnx_file)

    # Map from output arg to the producer of the output.
    output_to_node = {}
    for node in model.graph.node:
        for o in node.output:
            output_to_node[o] = node

    aten_ops = []
    python_ops = []
    memcpu_ops = []
    cast_ops = []
    msgs = []

    for node in model.graph.node:
        if "Memcpy" in node.op_type:
            memcpu_ops.append(f"{node.op_type} {node.name}")
        if node.op_type == "Cast":
            cast_ops.append(f"{node.name}")
        if node.op_type == "ATen":
            for attr in node.attribute:
                if attr.name == "operator":
                    aten_ops.append(f"{node.name}: {attr.s.decode('utf-8')}")
        if node.op_type == "PythonOp":
            for attr in node.attribute:
                if attr.name == "name":
                    python_ops.append(f"{node.name}: {attr.s.decode('utf-8')}")

        # Look for stand-alone Dropout node in *_execution_model_<mode>.onnx graph.
        # Examine whether it should be fused with surrounding Add ops into BiasDropout node.
        if node.op_type == "Dropout" and len(node.input) == 1:
            prev = output_to_node[node.input[0]]
            if prev.op_type == "Add":
                msgs.append(
                    f"Examine whether {node.name} should be fused with the leading {prev.name} op into BiasDropout node."
                )

        # Look for stand-alone Softmax node in *_execution_model_<mode>.onnx graph.
        # Examine whether it should be fused with the leading Add ops into BiasSoftmax node.
        if node.op_type == "Softmax" and len(node.input) == 1:
            prev = output_to_node[node.input[0]]
            if prev.op_type == "Add":
                msgs.append(
                    f"Examine whether {node.name} should be fused with the leading {prev.name} op into BiasSoftmax node."
                )

    if aten_ops:
        print("ATen op found:")
        for line in aten_ops:
            print(line)
        print(10 * "-")

    if python_ops:
        print("PythonOp found:")
        for line in python_ops:
            print(line)
        print(10 * "-")

    if memcpu_ops:
        print("Memcpu ops found:")
        for line in memcpu_ops:
            print(line)
        print(10 * "-")

    if cast_ops:
        print("Cast ops found:")
        for line in cast_ops:
            print(line)
        print(10 * "-")

    for line in msgs:
        print(line)


def main():
    process_file(args.onnx_file)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import pathlib


def check_all_delegates_have_unmanaged_function_pointer_attribute(file: pathlib.Path):
    """
    Check that all 'public delegate' declarations have a matching UnmanagedFunctionPointer attribute.
    :param file: C# source file to check.
    :return: Number of errors
    """

    print(f"Checking {file!s}")

    errors = 0
    line_num = 0
    with open(str(file.resolve(strict=True))) as f:
        prev_line = ""
        for line in f:
            line_num += 1

            # strip so it's easier to deal with commented out lines.
            line = line.strip()  # noqa
            if line.startswith("public delegate ") and not prev_line.startswith("[UnmanagedFunctionPointer"):
                errors += 1
                print(f"Line {line_num} is missing UnmanagedFunctionPointer attribute:\n\t{prev_line}\n\t{line}")

            prev_line = line

    return errors


def main():
    arg_parser = argparse.ArgumentParser(
        "Script to validate that the native delegates for the ONNX Runtime C# managed projects have the required "
        "attributes for iOS AOT. Paths are inferred from the script location."
        "Errors of this nature can only be detected at runtime, in a release build, of a Xamarin/MAUI app, "
        "on an actual iOS device. Due to that we take extra steps to identify problems early."
    )

    # no real args. just using this to provide description as help message
    _ = arg_parser.parse_args()

    # CI needs resolve() as __file__ is a relative path when the script is run there
    script_dir = pathlib.Path(__file__).resolve().parent
    csharp_root = script_dir.parent

    managed_dir = csharp_root / "src" / "Microsoft.ML.OnnxRuntime"
    native_methods = managed_dir / "NativeMethods.shared.cs"
    training_native_methods = managed_dir / "Training" / "NativeTrainingMethods.shared.cs"
    errors = check_all_delegates_have_unmanaged_function_pointer_attribute(native_methods)
    errors += check_all_delegates_have_unmanaged_function_pointer_attribute(training_native_methods)

    if errors:
        raise ValueError(f"{errors} errors were found. Please check output for specifics.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import os
import subprocess
import sys
import tempfile

from compare_results import Comparisons, compare_results_files

SCRIPT_DIR = os.path.realpath(os.path.dirname(__file__))


def parse_args():
    parser = argparse.ArgumentParser(description="Runs a BERT convergence test.")
    parser.add_argument("--binary_dir", required=True, help="Path to the ORT binary directory.")
    parser.add_argument("--training_data_root", required=True, help="Path to the training data root directory.")
    parser.add_argument("--model_root", required=True, help="Path to the model root directory.")
    parser.add_argument(
        "--gpu_sku",
        choices=["V100_16G", "MI100_32G"],
        default="V100_16G",
        required=False,
        help="GPU model (e.g. V100_16G, MI100_32G).",
    )
    return parser.parse_args()


def main():
    args = parse_args()

    with tempfile.TemporaryDirectory() as output_dir:
        convergence_test_output_path = os.path.join(output_dir, "convergence_test_out.csv")

        # run BERT training
        subprocess.run(  # noqa: PLW1510
            [
                os.path.join(args.binary_dir, "onnxruntime_training_bert"),
                "--model_name",
                os.path.join(
                    args.model_root,
                    "nv/bert-base/bert-base-uncased_L_12_H_768_A_12_V_30528_S_512_Dp_0.1_optimized_layer_norm_opset12",
                ),
                "--train_data_dir",
                os.path.join(args.training_data_root, "128/books_wiki_en_corpus/train"),
                "--test_data_dir",
                os.path.join(args.training_data_root, "128/books_wiki_en_corpus/test"),
                "--train_batch_size",
                "64",
                "--mode",
                "train",
                "--num_train_steps",
                "800",
                "--display_loss_steps",
                "5",
                "--optimizer",
                "adam",
                "--learning_rate",
                "5e-4",
                "--warmup_ratio",
                "0.1",
                "--warmup_mode",
                "Linear",
                "--gradient_accumulation_steps",
                "16",
                "--max_predictions_per_seq=20",
                "--use_mixed_precision",
                "--use_deterministic_compute",
                "--allreduce_in_fp16",
                "--lambda",
                "0",
                "--use_nccl",
                "--convergence_test_output_file",
                convergence_test_output_path,
                "--seed",
                "42",
                "--enable_grad_norm_clip=false",
            ]
        ).check_returncode()

        # reference data
        if args.gpu_sku == "MI100_32G":
            reference_csv = "bert_base.convergence.baseline.mi100.csv"
        elif args.gpu_sku == "V100_16G":
            reference_csv = "bert_base.convergence.baseline.csv"
        else:
            raise ValueError(f"Unrecognized gpu_sku {args.gpu_sku}")

        # verify output
        comparison_result = compare_results_files(
            expected_results_path=os.path.join(SCRIPT_DIR, "results", reference_csv),
            actual_results_path=convergence_test_output_path,
            field_comparisons={
                "step": Comparisons.eq(),
                "total_loss": Comparisons.float_le(1e-3),
                "mlm_loss": Comparisons.float_le(1e-3),
                "nsp_loss": Comparisons.float_le(1e-3),
            },
        )

        return 0 if comparison_result else 1


if __name__ == "__main__":
    sys.exit(main())



import os.path
import sys

import numpy as np
import onnx
from onnx import *  # noqa: F403


def find_node(graph_proto, op_type):
    nodes = []
    map_input_node = {}
    for node in graph_proto.node:
        if node.op_type == op_type:
            node_input = node.input[1] if op_type == "Div" or op_type == "Mul" else node.input[0]
            nodes.append(node)
            map_input_node[node_input] = node
    return nodes, map_input_node


def gen_attribute(key, value):
    attr = AttributeProto()  # noqa: F405
    attr.name = key
    attr.ints.extend(int(v) for v in value)
    attr.type = AttributeProto.INTS  # noqa: F405
    return attr


def main():
    if len(sys.argv) < 2:
        print("Please give model path...")
        return

    model_file_path = sys.argv[1]
    # model_file_path = os.path.dirname(sys.argv[1:])
    print("model_file_path: " + model_file_path)
    model_file_name = os.path.basename(model_file_path)
    print("model_file_name: " + model_file_name)

    new_model_file_path = model_file_path[:-5] + "_layer_norm.onnx"
    print(new_model_file_path)

    model_proto = onnx.load(model_file_path)
    # print(model_proto)

    graph_proto = model_proto.graph
    # print(graph_proto)
    # print(graph_proto.input)

    nodes_Div, map_input_Div = find_node(graph_proto, "Div")  # noqa: N806
    # print(map_input_Div)
    nodes_Sqrt, map_input_Sqrt = find_node(graph_proto, "Sqrt")  # noqa: N806
    # print(map_input_Sqrt)
    nodes_Add, map_input_Add = find_node(graph_proto, "Add")  # noqa: N806
    # print(map_input_Add)
    nodes_ReduceMean, map_input_ReduceMean = find_node(graph_proto, "ReduceMean")  # noqa: N806
    # print(map_input_ReduceMean)
    nodes_Pow, map_input_Pow = find_node(graph_proto, "Pow")  # noqa: N806
    # print(map_input_Pow)
    nodes_Mul, map_input_Mul = find_node(graph_proto, "Mul")  # noqa: N806

    # find right side Sub
    nodes_Sub = []  # noqa: N806
    map_input_Sub = {}  # noqa: N806
    for node in graph_proto.node:
        if node.op_type == "Sub":
            if node.output[0] in map_input_Pow:
                nodes_Sub.append(node)
                map_input_Sub[node.input[1]] = node
    # print(map_input_Sub)

    # find first ReduceMean
    first_ReduceMean = []  # noqa: N806
    first_ReduceMean_outputs = []  # noqa: N806
    for node in nodes_ReduceMean:
        if node.output[0] in map_input_Sub:
            first_ReduceMean.append(node)
            first_ReduceMean_outputs.append(node.output[0])
    # print(first_ReduceMean)

    # find constant node
    nodes_Constant = []  # noqa: N806
    map_output_Constant = {}  # noqa: N806
    for node in graph_proto.node:
        if node.op_type == "Constant":
            nodes_Constant.append(node)
            map_output_Constant[node.output[0]] = node
    # print(map_input_Sub)

    id = 0
    removed_nodes = []
    layer_norm_nodes = []
    # Replace with layer norm
    for node in first_ReduceMean:
        layer_norm_input = []
        layer_norm_output = []
        layer_norm_input.append(node.input[0])
        node_sub = map_input_Sub[node.output[0]]
        node_pow = map_input_Pow[node_sub.output[0]]
        node_reduce = map_input_ReduceMean[node_pow.output[0]]
        node_Add = map_input_Add[node_reduce.output[0]]  # noqa: N806
        node_Sqrt = map_input_Sqrt[node_Add.output[0]]  # noqa: N806
        node_Div = map_input_Div[node_Sqrt.output[0]]  # noqa: N806
        node_Mul = map_input_Mul[node_Div.output[0]]  # noqa: N806
        layer_norm_input.append(node_Mul.input[0])
        node_Add1 = map_input_Add[node_Mul.output[0]]  # noqa: N806
        layer_norm_input.append(node_Add1.input[1])
        removed_nodes.append(node)
        removed_nodes.append(node_sub)
        removed_nodes.append(node_pow)
        removed_nodes.append(node_reduce)
        removed_nodes.append(node_Add)
        removed_nodes.append(node_Sqrt)
        removed_nodes.append(node_Div)
        removed_nodes.append(node_Mul)
        removed_nodes.append(node_Add1)
        removed_nodes.append(map_output_Constant[node_pow.input[1]])
        # print(map_output_Constant[node_Add.input[1]])
        removed_nodes.append(map_output_Constant[node_Add.input[1]])
        layer_norm_output.append(node_Add1.output[0])
        id = id + 1
        layer_norm_output.append("saved_mean_" + str(id))
        id = id + 1
        layer_norm_output.append("saved_inv_std_var_" + str(id))
        layer_norm = helper.make_node(  # noqa: F405
            "LayerNormalization",
            layer_norm_input,
            layer_norm_output,
            "LayerNormalization_" + str(id),
            None,
            axis=node_reduce.attribute[0].ints[0],
            epsilon=9.999999960041972e-13,
        )
        layer_norm_nodes.append(layer_norm)

    # remove left side Subs
    for node in graph_proto.node:
        if node.op_type == "Sub":
            if node.input[1] in first_ReduceMean_outputs:
                removed_nodes.append(node)

    all_nodes = []
    for node in graph_proto.node:
        if node not in removed_nodes:
            all_nodes.append(node)

    for node in layer_norm_nodes:
        all_nodes.append(node)  # noqa: PERF402

    graph_proto.ClearField("node")
    graph_proto.node.extend(all_nodes)

    with open(new_model_file_path, "wb") as f:
        f.write(model_proto.SerializeToString())

    # Use ORT to verify the converted model. Notice that you must use python package from the
    # training branch because training requires some extra ops.
    import onnxruntime as ort

    # We convert model to accept variable-length batch size, so it can be any positive integer.
    batch = 3
    # This should match --max_seq_length when calling nv_run_pretraining.py.
    sq_length = 512
    # This should match vocab_size in bert_config.json in DeepLearningExamples/PyTorch/LanguageModeling/BERT.
    vocab_size = 30528

    # Create a fake data point.
    vocab_size = 30528  # It should match the value from BERT config file.
    input_ids = np.random.randint(low=0, high=vocab_size, size=(batch, sq_length), dtype=np.int64)
    segment_ids = np.random.randint(low=0, high=2, size=(batch, sq_length), dtype=np.int64)
    input_mask = np.ones((batch, sq_length), dtype=np.int64)

    # Do forward using the original model.
    sess = ort.InferenceSession(model_file_path, providers=ort.get_available_providers())
    result = sess.run(None, {"input1": input_ids, "input2": segment_ids, "input3": input_mask})

    # Do forward using the new model.
    new_sess = ort.InferenceSession(new_model_file_path, providers=ort.get_available_providers())
    new_result = new_sess.run(None, {"input1": input_ids, "input2": segment_ids, "input3": input_mask})

    # Compare the outcomes from the two models.
    print(np.linalg.norm(result[0] - new_result[0]))
    print(np.linalg.norm(result[1] - new_result[1]))


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import argparse
from dataclasses import dataclass

import numpy as np
from benchmark import BenchmarkOp, add_arguments


@dataclass
class OpParam:
    batch_size: int
    seq_len: int
    hidden_size: int
    data_type: type


class BenchmarkSkipLayerNorm(BenchmarkOp):
    def __init__(self, args):
        BenchmarkOp.__init__(self, args)

    @classmethod
    def create_inputs_outputs(cls, op_param):
        np.random.seed(0)
        input_data = np.random.rand(op_param.batch_size, op_param.seq_len, op_param.hidden_size).astype(
            op_param.data_type
        )
        skip = np.random.rand(op_param.batch_size, op_param.seq_len, op_param.hidden_size).astype(op_param.data_type)
        gamma = np.random.rand(op_param.hidden_size).astype(op_param.data_type)
        beta = np.random.rand(op_param.hidden_size).astype(op_param.data_type)
        bias = np.random.rand(op_param.hidden_size).astype(op_param.data_type)
        output_data = np.random.rand(op_param.batch_size, op_param.seq_len, op_param.hidden_size).astype(
            op_param.data_type
        )

        inputs = {
            "INPUT": input_data,
            "SKIP": skip,
            "GAMMA": gamma,
            "BETA": beta,
            "BIAS": bias,
        }
        outputs = {"return_val": output_data}

        return inputs, outputs

    def create_cases(self):
        model = (
            "models/skip_layer_norm_fp16.onnx" if self.args.precision == "fp16" else "models/skip_layer_norm_fp32.onnx"
        )
        data_type = np.float16 if self.args.precision == "fp16" else np.float32
        # bert-large
        op_param = OpParam(1, 384, 1024, data_type)
        self.add_case(op_param, model)

    @classmethod
    def case_profile(cls, op_param, time):
        profile = f"(batch seq_len hidden_size) = ({op_param.batch_size} {op_param.seq_len} {op_param.hidden_size}), {time:7.4f} ms"
        return profile


def main():
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    bm = BenchmarkSkipLayerNorm(args)
    bm.benchmark()


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os

from onnx import ModelProto, TensorProto, external_data_helper, load_model, numpy_helper
from parity_utilities import find_transformers_source

if find_transformers_source():
    from fusion_utils import NumpyHelper
else:
    from onnxruntime.transformers.fusion_utils import NumpyHelper


def fill_zeros_for_external_data(tensor: TensorProto):
    if tensor.HasField("raw_data"):  # already loaded
        return

    value = NumpyHelper.to_array(tensor, fill_zeros=True)
    zero_tensor = numpy_helper.from_array(value, name=tensor.name)
    tensor.raw_data = zero_tensor.raw_data


def fill_zeros_for_external_data_of_model(model: ModelProto):
    """Fill zeros value for external tensors of model

    Args:
        model (ModelProto): model to set external data
        base_dir (str): directory that contains external data
    """
    # TODO: support attribute tensor, which is rare in transformers model.
    for tensor in model.graph.initializer:
        if external_data_helper.uses_external_data(tensor):
            fill_zeros_for_external_data(tensor)
            # Change the state of tensors and remove external data
            tensor.data_location = TensorProto.DEFAULT
            del tensor.external_data[:]


def load_model_with_dummy_external_data(path: str) -> ModelProto:
    """Load model and fill zeros for initializers in external data.
       It helps in testing graph fusion of onnx model without external data.

    Args:
        path (str): onnx model path

    Returns:
        ModelProto: model with external data tensors filled with zero.
    """
    model = load_model(path, load_external_data=False)
    fill_zeros_for_external_data_of_model(model)
    return model


def get_test_data_path(sub_dir: str, file: str):
    relative_path = os.path.join(os.path.dirname(__file__), "test_data", sub_dir, file)
    if os.path.exists(relative_path):
        return relative_path
    return os.path.join(".", "transformers", "test_data", sub_dir, file)


def get_fusion_test_model(file: str):
    relative_path = os.path.join(os.path.dirname(__file__), "..", "..", "testdata", "transform", "fusion", file)
    if os.path.exists(relative_path):
        return relative_path
    return os.path.join(".", "testdata", "transform", "fusion", file)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# fused_adam.py
# This file has been adapted from microsoft/DeepSpeed

"""
Copyright 2020 The Microsoft DeepSpeed Team

Copyright NVIDIA/apex
This file is adapted from fused adam in NVIDIA/apex, commit a109f85
"""

from enum import IntEnum

import torch

from ._multi_tensor_apply import MultiTensorApply


class AdamWMode(IntEnum):
    ADAM_L2_REGULARIZATION = 0  # Adam with L2 regularization
    ADAMW_TRANSFORMERS = 1  # Adam with weight decay implemented to be equivalent to Transformers/AdamW
    ADAMW_TORCH = 2  # Adam with weight decay implemented to be equivalent to torch/AdamW


class FusedAdam(torch.optim.Optimizer):
    """Implements Adam algorithm.

    The algorithmic implementation is mathematically equivalent to
    `Transformers/AdamW <https://github.com/huggingface/transformers/blob/61f64262692ac7dc90e2e0bdeb7e79d9cd607a66/src/transformers/optimization.py#L349-L370>`_
    when adam_w_mode = 1 and `torch/Adam <https://github.com/pytorch/pytorch/blob/a217a62e73fd30b658743af8a69966f90327f018/torch/optim/adamw.py#L6>`_
    when adam_w_mode = 2

    Currently GPU-only.

    This version of fused Adam implements 2 fusions.

      * Fusion of the Adam update's elementwise operations
      * A multi-tensor apply launch that batches the elementwise updates applied to all the model's parameters into one or a few kernel launches.

    Adam was proposed in `Adam: A Method for Stochastic Optimization`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups.
        lr (float, optional): learning rate. (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square. (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability. (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        adam_w_mode (AdamWMode, optional): Apply L2 regularization or weight decay
            (AdamWMode.ADAM_L2_REGULARIZATION), decoupled weight decay with
            transformers/AdamW mathematical implementation (AdamWMode.ADAMW_TRANSFORMERS)
            or decoupled weight decay with transformers/AdamW implementation
            (AdamWMode.ADAMW_TORCH) (default: AdamWMode.ADAMW_TRANSFORMERS)
        set_grad_none (bool, optional): whether set grad to None when zero_grad()
            method is called. (default: True)
            PyTorch Adam has set_to_none parameter in zero_grad(), supporting that too

    .. _Adam - A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """

    def __init__(
        self,
        params,
        lr=1e-3,
        bias_correction=True,
        betas=(0.9, 0.999),
        eps=1e-6,
        adam_w_mode=AdamWMode.ADAMW_TRANSFORMERS,
        weight_decay=0.0,
        set_grad_none=True,
    ):
        # The FusedAdam implementation is mathematically equivalent to
        # transformers AdamW. The input arguments also have the same defaults.

        defaults = dict(lr=lr, bias_correction=bias_correction, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)
        self._adam_w_mode = adam_w_mode
        self._set_grad_none = set_grad_none

        # Skip buffer
        self._dummy_overflow_buf = torch.cuda.IntTensor([0])

        from onnxruntime.training.ortmodule.torch_cpp_extensions import fused_ops

        self._multi_tensor_adam = fused_ops.multi_tensor_adam
        self._multi_tensor_applier = MultiTensorApply(2048 * 32)
        self._TorchTensorVector = fused_ops.TorchTensorVector

    def zero_grad(self, set_to_none=True):
        if self._set_grad_none or set_to_none:
            for group in self.param_groups:
                for p in group["params"]:
                    p.grad = None
        else:
            super().zero_grad()

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.

        The remaining arguments are deprecated, and are only retained (for the moment) for error-checking purposes.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            bias_correction = 1 if group["bias_correction"] else 0
            beta1, beta2 = group["betas"]

            # assume same step across group now to simplify things
            # per parameter step can be easily support by making it tensor, or pass list into kernel
            if "step" in group:
                group["step"] += 1
            else:
                group["step"] = 1

            # create lists for multi-tensor apply
            g_16, p_16, m_16, v_16 = [], [], [], []
            g_32, p_32, m_32, v_32 = [], [], [], []

            for p in group["params"]:
                if p.grad is None:
                    continue
                if p.grad.data.is_sparse:
                    raise RuntimeError(
                        "FusedAdam does not support sparse gradients, please consider SparseAdam instead"
                    )

                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                if p.dtype == torch.float16:
                    g_16.append(p.grad.data)
                    p_16.append(p.data)
                    m_16.append(state["exp_avg"])
                    v_16.append(state["exp_avg_sq"])
                elif p.dtype == torch.float32:
                    g_32.append(p.grad.data)
                    p_32.append(p.data)
                    m_32.append(state["exp_avg"])
                    v_32.append(state["exp_avg_sq"])
                else:
                    raise RuntimeError("FusedAdam only support fp16 and fp32.")

            if len(g_16) > 0:
                self._multi_tensor_applier(
                    self._multi_tensor_adam,
                    self._dummy_overflow_buf,
                    [
                        self._TorchTensorVector(g_16),
                        self._TorchTensorVector(p_16),
                        self._TorchTensorVector(m_16),
                        self._TorchTensorVector(v_16),
                    ],
                    group["lr"],
                    beta1,
                    beta2,
                    group["eps"],
                    group["step"],
                    self._adam_w_mode,
                    bias_correction,
                    group["weight_decay"],
                )
            if len(g_32) > 0:
                self._multi_tensor_applier(
                    self._multi_tensor_adam,
                    self._dummy_overflow_buf,
                    [
                        self._TorchTensorVector(g_32),
                        self._TorchTensorVector(p_32),
                        self._TorchTensorVector(m_32),
                        self._TorchTensorVector(v_32),
                    ],
                    group["lr"],
                    beta1,
                    beta2,
                    group["eps"],
                    group["step"],
                    self._adam_w_mode,
                    bias_correction,
                    group["weight_decay"],
                )

        return loss



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import glob
import logging
import os
import re

logging.basicConfig(format="[%(levelname)s] - %(message)s", level=logging.DEBUG)
log = logging.getLogger()


def parse_args():
    parser = argparse.ArgumentParser(
        description="Find optimizers that involve operators which may need an update to the supported opset versions."
    )

    root_arg = parser.add_argument(
        "--ort-root", "-o", required=True, type=str, help="The root directory of the ONNX Runtime repository to search."
    )

    args = parser.parse_args()

    if not os.path.isdir(args.ort_root):
        raise argparse.ArgumentError(root_arg, f"{args.ort_root} is not a valid directory")

    return args


def get_call_args_from_file(filename: str, function_or_declaration: str) -> list[str]:
    """
    Search a file for all function calls or declarations that match the provided name.
    Requires both the opening '(' and closing ')' to be on the same line.
    Handles multiple calls being on the same line.
    """

    results = []
    with open(filename) as f:
        line_num = 0
        for line in f:
            for match in re.finditer(function_or_declaration, line):
                # check we have both the opening and closing brackets for the function call/declaration.
                # if we do we have all the arguments
                start = line.find("(", match.end())
                end = line.find(")", match.end())
                have_all_args = start != -1 and end != -1

                if have_all_args:
                    results.append(line[start + 1 : end])
                else:
                    # TODO: handle automatically by merging lines
                    log.error(
                        "Call/Declaration is split over multiple lines. Please check manually."
                        f"File:{filename} Line:{line_num}"
                    )
                    continue

            line_num += 1

    return results


def get_multiline_call_args_from_file(filename: str, function_or_declaration: str) -> list[str]:
    """
    Search a file for all function calls or declarations that match the provided name.
    Allows the opening '(' and closing ')' to be split across multiple lines.
    Supports a single call per line.
    """

    results = []
    with open(filename) as f:
        function_and_args = None

        for line in f:
            if not function_and_args:
                # look for new match
                start = line.find(function_or_declaration)
                if start != -1:
                    function_and_args = line[start:].strip()
            else:
                # append to existing line and look for closing ')'
                start = len(function_and_args)
                function_and_args += line.strip()

            if function_and_args:
                end = function_and_args.find(")", start)

                if end != -1:
                    start_args = function_and_args.find("(")
                    results.append(function_and_args[start_args + 1 : end])
                    function_and_args = None

    return results


def _add_if_newer(domain: str, op: str, opset: int, op_to_opset: dict[str, int]):
    key = domain + "." + op
    if key not in op_to_opset or op_to_opset[key] < opset:
        op_to_opset[key] = opset


def get_latest_ort_op_versions(root_dir):
    """Find the entries for the latest opset for each operator."""

    op_to_opset = {}
    files = [
        # for ONNX operators we use get_latest_onnx_op_versions
        # os.path.join(root_dir, "onnxruntime/core/providers/cpu/cpu_execution_provider.cc"),
        # for internal kernels we use the current registrations
        os.path.join(root_dir, "onnxruntime/contrib_ops/cpu/cpu_contrib_kernels.cc"),
        os.path.join(root_dir, "onnxruntime/contrib_ops/cuda/cuda_contrib_kernels.cc"),
    ]

    for file in files:
        # e.g. class ONNX_OPERATOR_KERNEL_CLASS_NAME(kCpuExecutionProvider, kOnnxDomain, 11, Clip);
        calls = get_multiline_call_args_from_file(file, "ONNX_OPERATOR_KERNEL_CLASS_NAME")
        for call in calls:
            args = call.split(",")
            domain = args[1].strip()
            opset = args[2].strip()
            op = args[3].strip()
            _add_if_newer(domain, op, int(opset), op_to_opset)

        # e.g. class ONNX_OPERATOR_TYPED_KERNEL_CLASS_NAME(kCpuExecutionProvider, kOnnxDomain, 11, float, ArgMax);
        calls = get_multiline_call_args_from_file(file, "ONNX_OPERATOR_TYPED_KERNEL_CLASS_NAME")
        for call in calls:
            args = call.split(",")
            domain = args[1].strip()
            opset = args[2].strip()
            op = args[4].strip()
            _add_if_newer(domain, op, int(opset), op_to_opset)

    return op_to_opset


def get_latest_onnx_op_versions(root_dir):
    """Get the latest versions of the ONNX operators from the ONNX headers."""

    op_to_opset = {}
    files = [
        # operators with domain of 'Onnx'
        os.path.join(root_dir, "cmake/external/onnx/onnx/defs/operator_sets.h"),
        # ML operators with domain of 'OnnxML'
        os.path.join(root_dir, "cmake/external/onnx/onnx/defs/operator_sets_ml.h"),
    ]

    for file in files:
        # e.g. fn(GetOpSchema<ONNX_OPERATOR_SET_SCHEMA_CLASS_NAME(Onnx, 17, LayerNormalization)>());
        #      fn(GetOpSchema<ONNX_OPERATOR_SET_SCHEMA_CLASS_NAME(OnnxML, 3, TreeEnsembleClassifier)>());
        calls = get_multiline_call_args_from_file(file, "ONNX_OPERATOR_SET_SCHEMA_CLASS_NAME")
        for call in calls:
            args = call.split(",")
            orig_domain = args[0].strip()
            # convert domain to the ORT constants
            domain = "kMLDomain" if orig_domain == "OnnxML" else "kOnnxDomain"
            opset = args[1].strip()
            op = args[2].strip()
            _add_if_newer(domain, op, int(opset), op_to_opset)

    return op_to_opset


def find_potential_issues(root_dir, op_to_opset):
    optimizer_dir = os.path.join(root_dir, "onnxruntime/core/optimizer")

    files = glob.glob(optimizer_dir + "/**/*.cc", recursive=True)
    files += glob.glob(optimizer_dir + "/**/*.h", recursive=True)

    for file in files:
        calls = get_call_args_from_file(file, "graph_utils::IsSupportedOptypeVersionAndDomain")
        for call in calls:
            # Need to handle multiple comma separated version numbers, and the optional domain argument.
            # e.g. IsSupportedOptypeVersionAndDomain(node, "MaxPool", {1, 8, 10})
            #      IsSupportedOptypeVersionAndDomain(node, "FusedConv", {1}, kMSDomain)
            args = call.split(",", 2)  # first 2 args are simple, remainder need custom processing
            op = args[1].strip()
            if not op.startswith('"') or not op.endswith('"'):
                log.error(f"Symbolic name of '{op}' found for op. Please check manually. File:{file}")
                continue

            versions_and_domain_arg = args[2]
            v1 = versions_and_domain_arg.find("{")
            v2 = versions_and_domain_arg.find("}")
            versions = versions_and_domain_arg[v1 + 1 : v2].split(",")
            last_version = versions[-1].strip()

            domain_arg_start = versions_and_domain_arg.find(",", v2)
            if domain_arg_start != -1:
                domain = versions_and_domain_arg[domain_arg_start + 1 :].strip()
            else:
                domain = "kOnnxDomain"

            op = domain + "." + op[1:-1]

            if op in op_to_opset:
                latest = op_to_opset[op]
                if int(latest) != int(last_version):
                    log.warning(
                        f"Newer opset found for {op}. Latest:{latest} Optimizer support ends at {last_version}. File:{file}"
                    )
            else:
                log.error(f"Failed to find version information for {op}. File:{file}")


if __name__ == "__main__":
    arguments = parse_args()
    ort_to_opset_map = get_latest_ort_op_versions(arguments.ort_root)
    onnx_op_to_opset_map = get_latest_onnx_op_versions(arguments.ort_root)

    # merge the two maps
    op_to_opset_map = {**ort_to_opset_map, **onnx_op_to_opset_map}

    find_potential_issues(arguments.ort_root, op_to_opset_map)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import argparse
from dataclasses import dataclass

import numpy as np
from benchmark import BenchmarkOp, add_arguments


@dataclass
class OpParam:
    n: int
    cout: int
    cin: int
    h: int
    w: int

    data_type: type


class BenchmarkNhwcConv(BenchmarkOp):
    def __init__(self, args):
        BenchmarkOp.__init__(self, args)

    @classmethod
    def create_inputs_outputs(cls, op_param):
        np.random.seed(0)
        input_data = np.random.rand(op_param.n, op_param.h, op_param.w, op_param.cin).astype(op_param.data_type)
        weight = np.random.rand(op_param.cout, 3, 3, op_param.cin).astype(op_param.data_type)
        bias = np.random.rand(op_param.cout).astype(op_param.data_type)
        output = np.random.rand(op_param.n, op_param.h, op_param.w, op_param.cout).astype(op_param.data_type)
        inputs = {"input": input_data, "weight": weight, "bias": bias}
        outputs = {"conv": output}
        return inputs, outputs

    def create_cases(self):
        # attributes of model : kernel_shape(3,3), group(1), pads(1,1), strides(1,1), dilations(1,1)
        model = "models/nhwcConv_fp16.onnx" if self.args.precision == "fp16" else "models/nhwcConv_fp32.onnx"
        data_type = np.float16 if self.args.precision == "fp16" else np.float32

        # change here to test your data shape
        self.add_case(OpParam(2, 320, 320, 64, 64, data_type), model)

    @classmethod
    def case_profile(cls, op_param, time):
        profile = f"( n cout cin h w ) = ( {op_param.n} {op_param.cout} {op_param.cin} {op_param.h} {op_param.w} ), {time * 1000:7.4f} us"
        return profile


def main():
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    bm = BenchmarkNhwcConv(args)
    bm.benchmark()


if __name__ == "__main__":
    main()



import onnx
from onnx import OperatorSetIdProto, TensorProto, helper

opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
onnxdomain.domain = ""  # The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
opsets.append(onnxdomain)

msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"

opsets.append(msdomain)
kwargs = {}
kwargs["opset_imports"] = opsets


def GenerateModel(model_name):  # noqa: N802
    nodes = [  # subgraph
        helper.make_node("Cast", ["A"], ["cast1"], "cast_1", to=11),
        helper.make_node("IsInf", ["cast1"], ["IsInf_out"], "is_inf"),
        helper.make_node("Cast", ["IsInf_out"], ["cast2"], "cast_2", to=7),
        helper.make_node("ReduceSum", ["cast2"], ["reduced"], "reduction", keepdims=0),
        helper.make_node("Greater", ["reduced", "one"], ["Y"], "output"),
    ]

    inputs = [  # inputs
        helper.make_tensor_value_info("A", TensorProto.FLOAT16, ["M", "K"]),
    ]

    initializers = [helper.make_tensor("one", TensorProto.INT64, [1], [1])]

    graph = helper.make_graph(
        nodes,
        "IsInfReduceSum",  # name
        inputs,
        [  # outputs
            helper.make_tensor_value_info("Y", TensorProto.BOOL, [1]),
        ],
        initializers,
    )

    model = helper.make_model(graph, **kwargs)
    onnx.save(model, model_name)


if __name__ == "__main__":
    GenerateModel("isinf_reducesum.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# This script opens an existing model in onnx format and attempts to
# move initializers from model.graph.initializer field to model.graph.sparse_initializer field
# and convert them into ONNX COO flat index format.
from __future__ import annotations

import argparse
import logging
import sys

import numpy as np
import onnx
from onnx import ModelProto, TensorProto, numpy_helper

logger = logging.getLogger(__name__)

real_types = {int(TensorProto.FLOAT), int(TensorProto.DOUBLE)}


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", required=True, type=str, help="input model path")
    parser.add_argument("--output", required=True, type=str, help="output model path")
    parser.add_argument(
        "--exclude", required=False, type=str, help="semicolon separated list of initializer names to exclude"
    )
    parser.add_argument("--tolerance", required=False, type=float, default=1e-6, help="FP absolute tolerance.")
    parser.add_argument(
        "--sparsity_threshold",
        required=False,
        type=float,
        default=0.5,
        help="convert to sparse initializers if sparsity is at least this much",
    )
    parser.add_argument("--verbose", required=False, action="store_true")
    parser.set_defaults(verbose=False)
    args = parser.parse_args()
    return args


def setup_logging(verbose):  # type: (bool)  -> None
    log_handler = logging.StreamHandler(sys.stdout)
    if verbose:
        log_handler.setFormatter(logging.Formatter("[%(filename)s:%(lineno)s - %(funcName)20s()] %(message)s"))
        logging_level = logging.DEBUG
    else:
        log_handler.setFormatter(logging.Formatter("%(filename)20s: %(message)s"))
        logging_level = logging.INFO
    log_handler.setLevel(logging_level)
    logger.addHandler(log_handler)
    logger.setLevel(logging_level)


def convert_tensor_to_sparse(tensor, sparsity_threshold, tolerance):  # type: (TensorProto, float, float) -> Tuple[SparseTensorProto, float]
    """returns a tuple of sparse_tensor and sparsity level"""
    values = []
    indices = []
    nnz_count = 0
    tensor_data = numpy_helper.to_array(tensor).flatten()
    data_len = len(tensor_data)
    if tensor_data.dtype in real_types:
        for index in range(data_len):
            el = tensor_data[index]
            if abs(el) <= tolerance:
                values.append(el)
                indices.append(index)
                nnz_count += 1
    else:
        for index in range(data_len):
            el = tensor_data[index]
            if el != 0:
                values.append(el)
                indices.append(index)
                nnz_count += 1

    sparsity = 1.0 - float(nnz_count) / data_len

    ind_data_type = TensorProto.INT8
    ind_dtype = np.int8
    ind_len = len(indices)
    max_indices_value = 0
    if ind_len > 0:
        max_indices_value = indices[-1]
        if max_indices_value <= np.iinfo(np.int8).max:
            ind_data_type = TensorProto.INT8
            ind_dtype = np.int8
        elif max_indices_value <= np.iinfo(np.int16).max:
            ind_data_type = TensorProto.INT16
            ind_dtype = np.int16
        elif max_indices_value <= np.iinfo(np.int32).max:
            ind_data_type = TensorProto.INT32
            ind_dtype = np.int32
        else:
            ind_data_type = TensorProto.INT64
            ind_dtype = np.int64

    logger.debug(
        f"initializer={tensor.name}, dtype={tensor_data.dtype}, \
                 data_len={data_len}, nnz={nnz_count}, sparsity={sparsity}, \
                 max_indices_value={max_indices_value}, sparse_indices_type={ind_dtype}"
    )

    if sparsity < sparsity_threshold:
        return (object(), sparsity)

    tensor_data_bytes = tensor_data.nbytes
    # create np array and cast data to the appropriate type
    np_values = np.array(values).astype(tensor_data.dtype)
    # create np array and cast data to the inferred index type
    np_indices = np.array(indices).astype(ind_dtype)
    total_sparse_bytes = np_values.nbytes + np_indices.nbytes

    logger.debug(
        f"initializer={tensor.name}, initializer_bytes={tensor_data_bytes}, \
                sparse_initializer_bytes={total_sparse_bytes}"
    )

    # This check is usually useful for sparsity_threshold=0.5 where much
    # depends on the size of the indices entries and the size of the original tensor.
    # Big dense tensors command larger indices data type and for large float32 tensors
    # int32 indices are often selected, thus we really want to guard against loosing
    # rather than winning.
    if tensor_data_bytes <= total_sparse_bytes:
        sparsity = 1.0 - float(tensor_data_bytes) / total_sparse_bytes
        logger.debug(f"initializer={tensor.name}, adjusted_sparsity={sparsity}")
        return (object(), sparsity)

    values_tensor = onnx.helper.make_tensor(tensor.name, tensor.data_type, [len(values)], np_values.tobytes(), raw=True)

    indicies_tensor = onnx.helper.make_tensor(
        tensor.name + "_indicies", ind_data_type, [ind_len], np_indices.tobytes(), raw=True
    )

    sparse_tensor = onnx.helper.make_sparse_tensor(values_tensor, indicies_tensor, tensor.dims)
    return (sparse_tensor, sparsity)


def convert_initializers(model, exclude_names, sparsity_threshold, tolerance):  # type: (ModelProto, List[str], float, float) -> None
    graph = model.graph
    converted_sparse = []
    remaining_initializers = []
    for initializer in graph.initializer:
        if initializer.name in exclude_names:
            logger.info(f"initializer={initializer.name} was excluded")
            continue
        if initializer.data_type == TensorProto.BOOL:
            logger.info(f"initializer={initializer.name} contains bool, not converted")
            remaining_initializers.append(initializer)
            continue
        sparse_tensor, sparsity = convert_tensor_to_sparse(initializer, sparsity_threshold, tolerance)
        if sparsity >= sparsity_threshold:
            logger.info(f"initializer={initializer.name} converted. sparsity={sparsity}")
            converted_sparse.append(sparse_tensor)
        else:
            remaining_initializers.append(initializer)
            logger.info(f"initializer={initializer.name} is not converted. sparsity={sparsity}")

    graph.sparse_initializer.extend(converted_sparse)
    del graph.initializer[:]
    graph.initializer.extend(remaining_initializers)


def main():
    args = parse_arguments()
    setup_logging(args.verbose)

    exclude_names = set() if args.exclude is None else set(args.exclude.split(";"))

    model = ModelProto()
    with open(args.input, "rb") as input_file:
        model.ParseFromString(input_file.read())

    convert_initializers(model, exclude_names, args.sparsity_threshold, args.tolerance)

    with open(args.output, "wb") as output_file:
        s = model.SerializeToString()
        output_file.write(s)


if __name__ == "__main__":
    main()



import numpy as np
import onnx
from onnx import TensorProto, helper, numpy_helper


def order_repeated_field(repeated_proto, key_name, order):
    order = list(order)
    repeated_proto.sort(key=lambda x: order.index(getattr(x, key_name)))


def make_node(op_type, inputs, outputs, name=None, doc_string=None, domain=None, **kwargs):
    node = helper.make_node(op_type, inputs, outputs, name, doc_string, domain, **kwargs)
    if doc_string == "":
        node.doc_string = ""
    order_repeated_field(node.attribute, "name", kwargs.keys())
    return node


def make_graph(*args, doc_string=None, **kwargs):
    graph = helper.make_graph(*args, doc_string=doc_string, **kwargs)
    if doc_string == "":
        graph.doc_string = ""
    return graph


test_graph = make_graph(
    name="test_graph",
    # model input of a sequence type to test IsSparseTensor issue
    inputs=[
        helper.make_tensor_sequence_value_info("seq_in", TensorProto.FLOAT, shape=None),
    ],
    outputs=[
        helper.make_tensor_value_info("still_has_elements", TensorProto.BOOL, shape=[]),
    ],
    initializer=[
        numpy_helper.from_array(np.array(0, dtype="int64"), name="i0"),
    ],
    nodes=[
        make_node("SequenceLength", inputs=["seq_in"], outputs=["seq_len"], name="get_seq_len"),
        make_node("Greater", inputs=["seq_len", "i0"], outputs=["has_elements"], name="get_has_elements"),
        # If node with one branch that has no nodes to test the allocation planner issue
        #   if sequence has elements:
        #     remove one
        #     output bool of whether it still has elements
        #   else:
        #     output false (gives us branch with no nodes)
        make_node(
            "If",
            name="test_if",
            inputs=["has_elements"],
            outputs=["still_has_elements"],
            then_branch=make_graph(
                name="then",
                inputs=[],
                outputs=[helper.make_tensor_value_info("then_bool_out", TensorProto.BOOL, shape=[])],
                nodes=[
                    make_node("SequenceErase", inputs=["seq_in", "i0"], outputs=["seq_less_one"]),
                    make_node("SequenceLength", inputs=["seq_less_one"], outputs=["new_seq_len"]),
                    make_node("Greater", inputs=["new_seq_len", "i0"], outputs=["then_bool_out"]),
                ],
            ),
            else_branch=make_graph(
                name="else",
                initializer=[numpy_helper.from_array(np.array(False, dtype="bool"), name="else_bool_out")],
                inputs=[],
                outputs=[helper.make_tensor_value_info("else_bool_out", TensorProto.BOOL, shape=[])],
                nodes=[],
            ),
        ),
    ],
)

# Graph with Sequence operations and an If node that has a subgraph with no nodes
model = helper.make_model(opset_imports=[helper.make_operatorsetid("ai.onnx", 14)], ir_version=7, graph=test_graph)

onnx.shape_inference.infer_shapes(model, strict_mode=True)
onnx.save(model, "ort_github_issue_17000.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import fileinput  # noqa: F401
import os
import sys  # noqa: F401

from setuptools import setup
from torch.utils import cpp_extension

filenames = [
    os.path.join(os.path.dirname(__file__), "fused_ops_frontend.cpp"),
    os.path.join(os.path.dirname(__file__), "multi_tensor_adam.cu"),
    os.path.join(os.path.dirname(__file__), "multi_tensor_scale_kernel.cu"),
    os.path.join(os.path.dirname(__file__), "multi_tensor_axpby_kernel.cu"),
    os.path.join(os.path.dirname(__file__), "multi_tensor_l2norm_kernel.cu"),
]

use_rocm = bool(os.environ["ONNXRUNTIME_ROCM_VERSION"])
extra_compile_args = {"cxx": ["-O3"]}
if not use_rocm:
    nvcc_extra_args = os.environ.get("ONNXRUNTIME_CUDA_NVCC_EXTRA_ARGS", "")
    if nvcc_extra_args:
        extra_compile_args.update({"nvcc": nvcc_extra_args.split(",")})

setup(
    name="fused_ops",
    ext_modules=[
        cpp_extension.CUDAExtension(name="fused_ops", sources=filenames, extra_compile_args=extra_compile_args)
    ],
    cmdclass={"build_ext": cpp_extension.BuildExtension},
)



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# pylint: disable=W0212,C0114,C0116

import copy
import sys
import unittest

import _test_helpers
import numpy as np
import torch
from numpy.testing import assert_almost_equal
from torch._C import _from_dlpack
from torch.utils.dlpack import from_dlpack

import onnxruntime as onnxrt
from onnxruntime.capi import _pybind_state as C
from onnxruntime.capi.onnxruntime_pybind11_state import OrtValue as C_OrtValue
from onnxruntime.capi.onnxruntime_pybind11_state import OrtValueVector
from onnxruntime.training.ortmodule import ORTModule, _utils

has_cuda = torch.cuda.is_available()


# Revoved code from _utils.


def C_ort_from_dlpack(dlpack):
    if hasattr(C, "ort_from_dlpack"):
        return C.ort_from_dlpack(dlpack)
    return from_dlpack(dlpack)


def _torch_tensor_from_dl_pack(dlpack, ortvalue, device):
    torch_tensor = from_dlpack(dlpack) if device.type != "ort" else C_ort_from_dlpack(dlpack)
    return torch_tensor.to(torch.bool) if ortvalue.data_type() == "tensor(bool)" else torch_tensor


def _ortvalue_to_torch_tensor(ortvalue, device):
    dlpack_tensor = ortvalue.to_dlpack()
    return _torch_tensor_from_dl_pack(dlpack_tensor, ortvalue, device)


def _ortvalues_to_torch_tensor(vect, device):
    return [_ortvalue_to_torch_tensor(ov, device) for ov in vect]


class TestOrtValue(unittest.TestCase):
    def testOrtValueDlPack_float32(self):
        numpy_arr_input = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=np.float32)
        ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(numpy_arr_input)
        self.assertEqual(numpy_arr_input.shape, tuple(ortvalue.shape()))
        ptr = ortvalue._ortvalue.data_ptr()

        dlp = ortvalue._ortvalue.to_dlpack()
        self.assertFalse(C.is_dlpack_uint8_tensor(dlp))
        ortvalue2 = C_OrtValue.from_dlpack(dlp, False)
        self.assertEqual(ptr, ortvalue2.data_ptr())
        new_array = ortvalue2.numpy()
        assert_almost_equal(numpy_arr_input, new_array)

        dlp = ortvalue._ortvalue.__dlpack__()
        self.assertFalse(C.is_dlpack_uint8_tensor(dlp))
        ortvalue2 = C_OrtValue.from_dlpack(dlp, False)
        self.assertEqual(ptr, ortvalue2.data_ptr())
        new_array = ortvalue2.numpy()
        assert_almost_equal(numpy_arr_input, new_array)

        device = ortvalue._ortvalue.__dlpack_device__()
        self.assertEqual((1, 0), device)

    def testOrtValueDlPack_bool(self):
        numpy_arr_input = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=bool)
        ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(numpy_arr_input)
        self.assertEqual(numpy_arr_input.shape, tuple(ortvalue.shape()))
        ptr = ortvalue._ortvalue.data_ptr()

        dlp = ortvalue._ortvalue.to_dlpack()
        self.assertTrue(C.is_dlpack_uint8_tensor(dlp))
        ortvalue2 = C_OrtValue.from_dlpack(dlp, True)
        self.assertEqual(ptr, ortvalue2.data_ptr())
        new_array = ortvalue2.numpy()
        assert_almost_equal(numpy_arr_input, new_array)

        dlp = ortvalue._ortvalue.__dlpack__()
        self.assertTrue(C.is_dlpack_uint8_tensor(dlp))
        ortvalue2 = C_OrtValue.from_dlpack(dlp, True)
        self.assertEqual(ptr, ortvalue2.data_ptr())
        new_array = ortvalue2.numpy()
        assert_almost_equal(numpy_arr_input, new_array)

        device = ortvalue._ortvalue.__dlpack_device__()
        self.assertEqual((1, 0), device)

    def testOrtValueVector_float32(self):
        narrays = [
            np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=np.float32),
            np.array([[6.0, 7.0], [8.0, 9.0], [1.0, 6.0]], dtype=np.float32),
        ]
        vect = OrtValueVector()
        for a in narrays:
            ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(a)
            vect.push_back(ortvalue._ortvalue)
        self.assertEqual(len(vect.bool_tensor_indices()), 0)
        self.assertEqual(len(vect), 2)
        for i, (ov, ar) in enumerate(zip(vect, narrays, strict=False)):
            ovar = ov.numpy()
            assert_almost_equal(ar, ovar)
            self.assertEqual(ov.element_type(), vect.element_type_at(i))

    def testOrtValueVector_bool(self):
        narrays = [
            np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=np.bool_),
            np.array([[6.0, 7.0], [8.0, 9.0], [1.0, 6.0]], dtype=np.bool_),
        ]
        vect = OrtValueVector()
        for a in narrays:
            ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(a)
            vect.push_back(ortvalue._ortvalue)
        self.assertEqual(vect.bool_tensor_indices(), [0, 1])
        self.assertEqual(len(vect), 2)
        for ov, ar in zip(vect, narrays, strict=False):
            ovar = ov.numpy()
            assert_almost_equal(ar, ovar)

    def OrtValueVectorDlPackOrtValue(self, my_to_tensor, tensor_type, device, dtype=np.float32):
        narrays = [
            np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=dtype),
            np.array([[6.0, 7.0], [8.0, 9.0], [1.0, 6.0]], dtype=dtype),
        ]
        vect = OrtValueVector()
        ptr = []
        for a in narrays:
            ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(a, device)
            vect.push_back(ortvalue._ortvalue)
            ptr.append(ortvalue.data_ptr())
        self.assertEqual(len(vect), 2)

        converted_values = vect.to_dlpacks(my_to_tensor)
        self.assertEqual(len(converted_values), len(vect))
        if my_to_tensor is None:
            self.assertIn("PyCapsule", str(type(converted_values[0])))
            converted_values = [C_OrtValue.from_dlpack(o, False) for o in converted_values]
        else:
            assert all(isinstance(v, tensor_type) for v in converted_values)

        # We make sure the function does not leak any python object.
        cf = [sys.getrefcount(o) for o in converted_values]
        dummy = [np.array([[0, 1]]), {"a": 3}]
        cf2 = [sys.getrefcount(o) for o in dummy]
        self.assertEqual(cf, cf2)  # it should be [3, 3]

        ptr2 = []
        for av1, v2 in zip(narrays, converted_values, strict=False):
            ptr2.append(v2.data_ptr())
            if hasattr(v2, "cpu"):
                av2 = v2.cpu().numpy()
            else:
                av2 = v2.numpy()
            assert_almost_equal(av1, av2)
        self.assertEqual(ptr, ptr2)

    def testOrtValueVectorDlPackOrtValue_cpu(self):
        def my_to_tensor(dlpack_structure):
            return C_OrtValue.from_dlpack(dlpack_structure, False)

        self.OrtValueVectorDlPackOrtValue(my_to_tensor, C_OrtValue, "cpu")

    def testOrtValueVectorDlPackTorch_cpu(self):
        def my_to_tensor(dlpack_structure):
            return from_dlpack(dlpack_structure)

        self.OrtValueVectorDlPackOrtValue(my_to_tensor, torch.Tensor, "cpu")

    def testOrtValueVectorDlPack_Torch_cpu(self):
        def my_to_tensor(dlpack_structure):
            return _from_dlpack(dlpack_structure)

        self.OrtValueVectorDlPackOrtValue(my_to_tensor, torch.Tensor, "cpu")

    def testOrtValueVectorDlPackNone_cpu(self):
        self.OrtValueVectorDlPackOrtValue(None, None, "cpu")

    @unittest.skipIf(not has_cuda, reason="No CUDA availabled.")
    def testOrtValueVectorDlPackOrtValue_cuda(self):
        def my_to_tensor(dlpack_structure):
            return C_OrtValue.from_dlpack(dlpack_structure, False)

        self.OrtValueVectorDlPackOrtValue(my_to_tensor, C_OrtValue, "cuda")

    @unittest.skipIf(not has_cuda, reason="No CUDA availabled.")
    def testOrtValueVectorDlPackTorch_cuda(self):
        def my_to_tensor(dlpack_structure):
            return from_dlpack(dlpack_structure)

        self.OrtValueVectorDlPackOrtValue(my_to_tensor, torch.Tensor, "cuda")

    @unittest.skipIf(not has_cuda, reason="No CUDA availabled.")
    def testOrtValueVectorDlPack_Torch_cuda(self):
        def my_to_tensor(dlpack_structure):
            return _from_dlpack(dlpack_structure)

        self.OrtValueVectorDlPackOrtValue(my_to_tensor, torch.Tensor, "cuda")

    @unittest.skipIf(not has_cuda, reason="No CUDA availabled.")
    def testOrtValueVectorDlPackNone_cuda(self):
        self.OrtValueVectorDlPackOrtValue(None, None, "cuda")

    def test_ortmodule_dlpack(self):
        class NeuralNetTanh(torch.nn.Module):
            def __init__(self, input_size, hidden_size, num_classes):
                super().__init__()

                self.fc1 = torch.nn.Linear(input_size, hidden_size)
                self.tanh = torch.nn.Tanh()

            def forward(self, input1):
                out = self.fc1(input1)
                out = self.tanh(out)
                return out

        def run_step(model, x):
            prediction = model(x)
            loss = prediction.sum()
            loss.backward()
            return prediction, loss

        N, D_in, H, D_out = 120, 1536, 500, 1536  # noqa: N806
        pt_model = NeuralNetTanh(D_in, H, D_out)
        ort_model = ORTModule(copy.deepcopy(pt_model))

        for _step in range(10):
            pt_x = torch.randn(N, D_in, device="cpu", requires_grad=True)
            ort_x = copy.deepcopy(pt_x)
            ort_prediction, ort_loss = run_step(ort_model, ort_x)
            pt_prediction, pt_loss = run_step(pt_model, pt_x)
            _test_helpers.assert_values_are_close(ort_prediction, pt_prediction, atol=1e-4)
            _test_helpers.assert_values_are_close(ort_x.grad, pt_x.grad)
            _test_helpers.assert_values_are_close(ort_loss, pt_loss, atol=1e-4)

    def test_bool_input_and_output(self):
        class NeuralNetBoolInputOutput(torch.nn.Module):
            def __init__(self, input_size, num_classes):
                super().__init__()
                self.fc = torch.nn.Linear(input_size, num_classes)
                self.relu = torch.nn.ReLU()

            def forward(self, condition, x1, x2):
                out1 = self.relu(self.fc(torch.where(condition, x1, x2)))
                out2 = torch.tensor(out1).to(torch.bool)
                return out1, out2

        device = "cpu"
        N, D_in, D_out = 8, 16, 2  # noqa: N806
        model = NeuralNetBoolInputOutput(D_in, D_out).to(device)
        model = ORTModule(model)
        condition = torch.randint(2, (N, D_in), dtype=torch.bool, device=device)
        x1 = torch.randn(N, D_in, device=device)
        x2 = torch.randn(N, D_in, device=device)
        y1, y2 = model(condition, x1, x2)

        assert y1 is not None
        assert y2 is not None and y2.dtype == torch.bool

    def _ortvalues_to_torch_tensor_ortvaluevector(self, device, tensor_type, new_impl, dtype=np.float32):
        narrays = [
            np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=dtype),
            np.array([[6.0, 7.0], [8.0, 9.0], [1.0, 6.0]], dtype=dtype),
        ]
        vect = OrtValueVector()
        ptr = []
        for a in narrays:
            ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(a, device.type if device.type != "ort" else "cpu")
            vect.push_back(ortvalue._ortvalue)
            ptr.append(ortvalue.data_ptr())
        self.assertEqual(len(vect), 2)
        if new_impl == "list":
            raise AssertionError("Conversion from list to torch is not supported anymore.")
        if new_impl:
            tensors = _utils._ortvalues_to_torch_tensor(vect, device)
        else:
            tensors = _ortvalues_to_torch_tensor(vect, device)
        self.assertEqual(len(tensors), len(vect))
        for t in tensors:
            assert isinstance(t, torch.Tensor)
        self.assertEqual(ptr, [t.data_ptr() for t in tensors])
        assert all(isinstance(v, tensor_type) for v in tensors)

    def test_ortvalues_to_torch_tensor_ortvaluevector_cpu_new(self):
        device = torch.device("cpu")
        self._ortvalues_to_torch_tensor_ortvaluevector(device, torch.Tensor, True)

    def test_ortvalues_to_torch_tensor_ortvaluevector_cpu_old(self):
        device = torch.device("cpu")
        self._ortvalues_to_torch_tensor_ortvaluevector(device, torch.Tensor, False)

    def test_ortvalues_to_torch_tensor_ortvaluevector_ort_new(self):
        device = torch.device("ort")
        if hasattr(C, "ort_from_dlpack"):
            self._ortvalues_to_torch_tensor_ortvaluevector(device, torch.Tensor, True)
        else:
            with self.assertRaises(AttributeError):
                self._ortvalues_to_torch_tensor_ortvaluevector(device, torch.Tensor, True)

    def test_ortvalues_to_torch_tensor_ortvaluevector_ort_old(self):
        device = torch.device("ort")
        self._ortvalues_to_torch_tensor_ortvaluevector(device, torch.Tensor, False)

    @unittest.skipIf(not has_cuda, reason="No CUDA availabled.")
    def test_ortvalues_to_torch_tensor_ortvaluevector_cuda_new(self):
        device = torch.device("cuda:0")
        self._ortvalues_to_torch_tensor_ortvaluevector(device, torch.Tensor, True)

    @unittest.skipIf(not has_cuda, reason="No CUDA availabled.")
    def test_ortvalues_to_torch_tensor_ortvaluevector_cuda_old(self):
        device = torch.device("cuda:0")
        self._ortvalues_to_torch_tensor_ortvaluevector(device, torch.Tensor, False)

    def _ortvalues_to_torch_tensor_list(self, device, tensor_type, new_impl):
        narrays = [
            np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=np.float32),
            np.array([[6.0, 7.0], [8.0, 9.0], [1.0, 6.0]], dtype=np.float32),
        ]
        vect = C.OrtValueVector()
        vect.reserve(len(narrays))
        ptr = []
        for a in narrays:
            ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(a, device.type if device.type != "ort" else "cpu")
            vect.push_back(ortvalue._ortvalue)
            ptr.append(ortvalue.data_ptr())
        self.assertEqual(len(vect), 2)
        if new_impl:
            tensors = _utils._ortvalues_to_torch_tensor(vect, device)
        else:
            tensors = _ortvalues_to_torch_tensor(vect, device)
        self.assertEqual(len(tensors), len(vect))
        self.assertEqual(ptr, [t.data_ptr() for t in tensors])
        assert all(isinstance(v, tensor_type) for v in tensors)

    def test_ortvalues_to_torch_tensor_list_cpu_new(self):
        device = torch.device("cpu")
        self._ortvalues_to_torch_tensor_list(device, torch.Tensor, True)

    def test_ortvalues_to_torch_tensor_list_cpu_old(self):
        device = torch.device("cpu")
        self._ortvalues_to_torch_tensor_list(device, torch.Tensor, False)

    def test_ortvalues_to_torch_tensor_list_ort_new(self):
        device = torch.device("ort")
        if hasattr(C, "ort_from_dlpack"):
            self._ortvalues_to_torch_tensor_list(device, torch.Tensor, True)
        else:
            with self.assertRaises(AttributeError):
                self._ortvalues_to_torch_tensor_list(device, torch.Tensor, True)

    def test_ortvalues_to_torch_tensor_list_ort_old(self):
        device = torch.device("ort")
        self._ortvalues_to_torch_tensor_list(device, torch.Tensor, False)

    @unittest.skipIf(not has_cuda, reason="No CUDA availabled.")
    def test_ortvalues_to_torch_tensor_list_cuda_new(self):
        device = torch.device("cuda:0")
        self._ortvalues_to_torch_tensor_list(device, torch.Tensor, True)

    @unittest.skipIf(not has_cuda, reason="No CUDA availabled.")
    def test_ortvalues_to_torch_tensor_list_cuda_old(self):
        device = torch.device("cuda:0")
        self._ortvalues_to_torch_tensor_list(device, torch.Tensor, False)

    def test_element_type(self):
        values = {
            np.bool_: ("tensor(bool)", 9),
            np.int8: ("tensor(int8)", 3),
            np.uint8: ("tensor(uint8)", 2),
            np.int16: ("tensor(int16)", 5),
            np.uint16: ("tensor(uint16)", 4),
            np.int32: ("tensor(int32)", 6),
            np.uint32: ("tensor(uint32)", 12),
            np.int64: ("tensor(int64)", 7),
            np.uint64: ("tensor(uint64)", 13),
            np.float16: ("tensor(float16)", 10),
            np.float32: ("tensor(float)", 1),
            np.float64: ("tensor(double)", 11),
        }
        for dt, expected in values.items():
            numpy_arr_input = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=dt)
            ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(numpy_arr_input)
            stype = ortvalue.data_type()
            self.assertIn(stype, expected)
            proto_type = ortvalue.element_type()
            self.assertIn(proto_type, expected)


if __name__ == "__main__":
    unittest.main()
